"""OT3 Hardware Controller Backend."""

from __future__ import annotations
import asyncio
from contextlib import asynccontextmanager
from functools import wraps
import logging
from copy import deepcopy
from numpy import isclose
from typing import (
    Any,
    Awaitable,
    Callable,
    Dict,
    List,
    Optional,
    Tuple,
    Sequence,
    AsyncIterator,
    cast,
    Set,
    TypeVar,
    Iterator,
    KeysView,
    Union,
    Mapping,
)
from opentrons.config.types import OT3Config, GantryLoad
from opentrons.config import gripper_config
from .ot3utils import (
    axis_convert,
    create_move_group,
    axis_to_node,
    get_current_settings,
    create_home_groups,
    node_to_axis,
    sensor_node_for_mount,
    sensor_node_for_pipette,
    sensor_id_for_instrument,
    create_gripper_jaw_grip_group,
    create_gripper_jaw_home_group,
    create_gripper_jaw_hold_group,
    create_tip_action_group,
    create_tip_motor_home_group,
    motor_nodes,
    LIMIT_SWITCH_OVERTRAVEL_DISTANCE,
    map_pipette_type_to_sensor_id,
    moving_pipettes_in_move_group,
    gripper_jaw_state_from_fw,
    get_system_constraints,
    get_system_constraints_for_calibration,
    get_system_constraints_for_plunger_acceleration,
)
from .tip_presence_manager import TipPresenceManager

try:
    import aionotify  # type: ignore[import-untyped]
except (OSError, ModuleNotFoundError):
    aionotify = None


from opentrons_hardware.drivers import SystemDrivers
from opentrons_hardware.drivers.can_bus import CanMessenger, DriverSettings
from opentrons_hardware.drivers.can_bus.abstract_driver import AbstractCanDriver
from opentrons_hardware.drivers.can_bus.build import build_driver
from opentrons_hardware.drivers.binary_usb import (
    BinaryMessenger,
    SerialUsbDriver,
    build_rear_panel_driver,
)
from opentrons_hardware.drivers.eeprom import EEPROMDriver, EEPROMData
from opentrons_hardware.hardware_control.move_group_runner import MoveGroupRunner
from opentrons_hardware.hardware_control.motion_planning import (
    MoveManager,
    MoveTarget,
    ZeroLengthMoveError,
)
from opentrons_hardware.hardware_control.estop.detector import (
    EstopDetector,
)

from opentrons.hardware_control.backends.estop_state import EstopStateMachine

from opentrons_hardware.hardware_control.motor_enable_disable import (
    set_enable_motor,
    set_disable_motor,
    set_enable_tip_motor,
    set_disable_tip_motor,
    get_motor_enabled,
)
from opentrons_hardware.hardware_control.motor_position_status import (
    get_motor_position,
    update_motor_position_estimation,
)
from opentrons_hardware.hardware_control.limit_switches import get_limit_switches
from opentrons_hardware.hardware_control.current_settings import (
    set_run_current,
    set_hold_current,
    set_currents,
)
from opentrons_hardware.firmware_bindings.constants import (
    NodeId,
    PipetteName as FirmwarePipetteName,
    ErrorCode,
    SensorId,
)
from opentrons_hardware.firmware_bindings.messages.message_definitions import (
    StopRequest,
)
from opentrons_hardware.firmware_bindings.messages.payloads import EmptyPayload
from opentrons_hardware.hardware_control import status_bar

from opentrons_hardware.firmware_bindings.binary_constants import BinaryMessageId
from opentrons_hardware.firmware_bindings.messages.binary_message_definitions import (
    BinaryMessageDefinition,
    DoorSwitchStateInfo,
)
from opentrons_hardware.firmware_update import FirmwareUpdate
from opentrons_hardware.hardware_control import network, tools

from opentrons.hardware_control.module_control import AttachedModulesControl
from opentrons.hardware_control.types import (
    BoardRevision,
    Axis,
    AionotifyEvent,
    OT3Mount,
    OT3AxisMap,
    OT3AxisKind,
    CurrentConfig,
    MotorStatus,
    InstrumentProbeType,
    UpdateStatus,
    DoorState,
    SubSystemState,
    SubSystem,
    TipStateType,
    GripperJawState,
    HardwareFeatureFlags,
    EstopOverallStatus,
    EstopAttachLocation,
    EstopState,
    HardwareEventHandler,
    HardwareEventUnsubscriber,
    PipetteSensorId,
    PipetteSensorType,
    PipetteSensorData,
    PipetteSensorResponseQueue,
)
from opentrons.hardware_control.errors import (
    InvalidPipetteName,
    InvalidPipetteModel,
)
from opentrons_hardware.hardware_control.motion import (
    MoveStopCondition,
    MoveGroup,
)
from opentrons_hardware.hardware_control.types import (
    NodeMap,
    MotorPositionStatus,
    MoveCompleteAck,
)
from opentrons_hardware.hardware_control.tools import types as ohc_tool_types

from opentrons_hardware.hardware_control.tool_sensors import (
    capacitive_probe,
    capacitive_pass,
    liquid_probe,
    check_overpressure,
    grab_pressure,
)
from opentrons_hardware.hardware_control.rear_panel_settings import (
    get_door_state,
    set_deck_light,
    get_deck_light_state,
)
from opentrons_hardware.hardware_control.gripper_settings import (
    get_gripper_jaw_state,
)
from opentrons_hardware.hardware_control.hepa_uv_settings import (
    set_hepa_fan_state as set_hepa_fan_state_fw,
    get_hepa_fan_state as get_hepa_fan_state_fw,
    set_hepa_uv_state as set_hepa_uv_state_fw,
    get_hepa_uv_state as get_hepa_uv_state_fw,
)

from opentrons_hardware.drivers.gpio import OT3GPIO, RemoteOT3GPIO
from opentrons_shared_data.pipette.types import PipetteName
from opentrons_shared_data.pipette import (
    pipette_load_name_conversions as pipette_load_name,
    load_data as load_pipette_data,
)
from opentrons_shared_data.gripper.gripper_definition import GripForceProfile

from opentrons_shared_data.errors.exceptions import (
    EStopActivatedError,
    EStopNotPresentError,
    PipetteOverpressureError,
    FirmwareUpdateRequiredError,
    FailedGripperPickupError,
    PipetteLiquidNotFoundError,
    CommunicationError,
    PythonException,
    UnsupportedHardwareCommand,
)

from .subsystem_manager import SubsystemManager

from ..dev_types import (
    AttachedPipette,
    AttachedGripper,
    OT3AttachedInstruments,
)
from ..types import HepaFanState, HepaUVState, StatusBarState

from .types import HWStopCondition
from .flex_protocol import FlexBackend
from .status_bar_state import StatusBarStateController
from opentrons_hardware.sensors.sensor_types import (
    EnvironmentSensor,
    CapacitiveSensor,
    PressureSensor,
)
from opentrons_hardware.sensors.types import SensorDataType, EnvironmentSensorDataType
from opentrons_hardware.sensors.sensor_driver import SensorDriver
from opentrons_hardware.sensors.utils import send_evo_dispense_count_increase

log = logging.getLogger(__name__)

MapPayload = TypeVar("MapPayload")
Wrapped = TypeVar("Wrapped", bound=Callable[..., Awaitable[Any]])


def requires_update(func: Wrapped) -> Wrapped:
    """Decorator that raises FirmwareUpdateRequiredError if the update_required flag is set."""

    @wraps(func)
    async def wrapper(self: Any, *args: Any, **kwargs: Any) -> Any:
        if self.update_required and self.initialized:
            raise FirmwareUpdateRequiredError(
                func.__name__,
                self.subsystems_to_update,
            )
        return await func(self, *args, **kwargs)

    return cast(Wrapped, wrapper)


def requires_estop(func: Wrapped) -> Wrapped:
    """Decorator that raises an exception if the Estop is engaged."""

    @wraps(func)
    async def wrapper(self: OT3Controller, *args: Any, **kwargs: Any) -> Any:
        state = self._estop_state_machine.state
        if state == EstopState.NOT_PRESENT and self._feature_flags.require_estop:
            raise EStopNotPresentError(
                message="An Estop must be plugged in to move the robot."
            )
        if state == EstopState.LOGICALLY_ENGAGED:
            raise EStopActivatedError(
                message="Estop must be acknowledged and cleared to move the robot."
            )
        if state == EstopState.PHYSICALLY_ENGAGED:
            raise EStopActivatedError(
                message="Estop is currently engaged, robot cannot move."
            )
        return await func(self, *args, **kwargs)

    return cast(Wrapped, wrapper)


class OT3Controller(FlexBackend):
    """OT3 Hardware Controller Backend."""

    _initialized: bool
    _messenger: CanMessenger
    _usb_messenger: Optional[BinaryMessenger]
    _position: Dict[NodeId, float]
    _encoder_position: Dict[NodeId, float]
    _motor_status: Dict[NodeId, MotorStatus]
    _subsystem_manager: SubsystemManager
    _engaged_axes: OT3AxisMap[bool]

    @classmethod
    async def build(
        cls,
        config: OT3Config,
        use_usb_bus: bool = False,
        check_updates: bool = True,
        feature_flags: Optional[HardwareFeatureFlags] = None,
    ) -> OT3Controller:
        """Create the OT3Controller instance.

        Args:
            config: Robot configuration

        Returns:
            Instance.
        """
        driver = await build_driver(DriverSettings())
        usb_driver = None
        if use_usb_bus:
            try:
                usb_driver = await build_rear_panel_driver()
            except IOError as e:
                log.error(
                    "No rear panel device found, probably an EVT bot, disable rearPanelIntegration feature flag if it is"
                )
                raise e
        inst = cls(
            config,
            driver=driver,
            usb_driver=usb_driver,
            check_updates=check_updates,
            feature_flags=feature_flags,
        )
        await inst._subsystem_manager.start()
        return inst

    def __init__(
        self,
        config: OT3Config,
        driver: AbstractCanDriver,
        usb_driver: Optional[SerialUsbDriver] = None,
        eeprom_driver: Optional[EEPROMDriver] = None,
        check_updates: bool = True,
        feature_flags: Optional[HardwareFeatureFlags] = None,
    ) -> None:
        """Construct.

        Args:
            config: Robot configuration
            driver: The Can Driver
        """
        self._configuration = config
        self._module_controls: Optional[AttachedModulesControl] = None
        self._messenger = CanMessenger(driver=driver)
        self._messenger.start()
        self._drivers = self._build_system_hardware(
            self._messenger, usb_driver, eeprom_driver
        )
        self._feature_flags = feature_flags or HardwareFeatureFlags()
        self._usb_messenger = self._drivers.usb_messenger
        self._gpio_dev = self._drivers.gpio_dev
        self._subsystem_manager = SubsystemManager(
            self._messenger,
            self._usb_messenger,
            tools.detector.ToolDetector(self._messenger),
            network.NetworkInfo(self._messenger, self._usb_messenger),
            FirmwareUpdate(),
        )
        self._estop_detector: Optional[EstopDetector] = None
        self._estop_state_machine = EstopStateMachine(detector=None)
        self._position = self._get_home_position()
        self._gear_motor_position: Dict[NodeId, float] = {}
        self._encoder_position = self._get_home_position()
        self._motor_status = {}
        self._engaged_axes = {}
        self._check_updates = check_updates
        self._initialized = False
        self._status_bar = status_bar.StatusBar(messenger=self._usb_messenger)
        self._status_bar_controller = StatusBarStateController(self._status_bar)

        try:
            self._event_watcher = self._build_event_watcher()
        except AttributeError:
            log.warning(
                "Failed to initiate aionotify, cannot watch modules "
                "or door, likely because not running on linux"
            )
        self._current_settings: Optional[OT3AxisMap[CurrentConfig]] = None
        self._tip_presence_manager = TipPresenceManager(self._messenger)
        self._move_manager = MoveManager(
            constraints=get_system_constraints(
                self._configuration.motion_settings, GantryLoad.LOW_THROUGHPUT
            )
        )
        self._pressure_sensor_available: Dict[NodeId, bool] = {}

    @asynccontextmanager
    async def restore_system_constraints(self) -> AsyncIterator[None]:
        old_system_constraints = deepcopy(self._move_manager.get_constraints())
        try:
            yield
        finally:
            self._move_manager.update_constraints(old_system_constraints)
            log.debug(f"Restore previous system constraints: {old_system_constraints}")

    @asynccontextmanager
    async def grab_pressure(
        self, channels: int, mount: OT3Mount
    ) -> AsyncIterator[None]:
        tool = axis_to_node(Axis.of_main_tool_actuator(mount))
        async with grab_pressure(channels, tool, self._messenger):
            yield

    def set_pressure_sensor_available(
        self, pipette_axis: Axis, available: bool
    ) -> None:
        pip_node = axis_to_node(pipette_axis)
        self._pressure_sensor_available[pip_node] = available

    def get_pressure_sensor_available(self, pipette_axis: Axis) -> bool:
        pip_node = axis_to_node(pipette_axis)
        return self._pressure_sensor_available[pip_node]

    def update_constraints_for_calibration_with_gantry_load(
        self,
        gantry_load: GantryLoad,
    ) -> None:
        self._move_manager.update_constraints(
            get_system_constraints_for_calibration(
                self._configuration.motion_settings, gantry_load
            )
        )
        log.debug(
            f"Set system constraints for calibration: {self._move_manager.get_constraints()}"
        )

    def update_constraints_for_gantry_load(self, gantry_load: GantryLoad) -> None:
        self._move_manager.update_constraints(
            get_system_constraints(self._configuration.motion_settings, gantry_load)
        )

    def update_constraints_for_plunger_acceleration(
        self,
        mount: OT3Mount,
        acceleration: float,
        gantry_load: GantryLoad,
        high_speed_pipette: bool = False,
    ) -> None:
        new_constraints = get_system_constraints_for_plunger_acceleration(
            self._configuration.motion_settings,
            gantry_load,
            mount,
            acceleration,
            high_speed_pipette,
        )
        self._move_manager.update_constraints(new_constraints)

    async def get_serial_number(self) -> Optional[str]:
        if not self.initialized:
            return None
        return self.eeprom_data.serial_number

    @property
    def initialized(self) -> bool:
        """True when the hardware controller has initialized and is ready."""
        return self._initialized

    @initialized.setter
    def initialized(self, value: bool) -> None:
        self._initialized = value

    @property
    def subsystems(self) -> Dict[SubSystem, SubSystemState]:
        return self._subsystem_manager.subsystems

    @property
    def fw_version(self) -> Dict[SubSystem, int]:
        """Get the firmware version."""
        return {
            subsystem: info.current_fw_version
            for subsystem, info in self.subsystems.items()
        }

    @property
    def eeprom_driver(self) -> EEPROMDriver:
        """The eeprom driver interface."""
        return self._drivers.eeprom

    @property
    def eeprom_data(self) -> EEPROMData:
        """Get the data on the eeprom."""
        return self._drivers.eeprom.data

    @property
    def update_required(self) -> bool:
        return self._subsystem_manager.update_required and self._check_updates

    @property
    def subsystems_to_update(self) -> List[SubSystem]:
        return self._subsystem_manager.subsystems_to_update

    @staticmethod
    def _build_system_hardware(
        can_messenger: CanMessenger,
        usb_driver: Optional[SerialUsbDriver],
        eeprom_driver: Optional[EEPROMDriver],
    ) -> SystemDrivers:
        gpio = OT3GPIO("hardware_control")
        eeprom_driver = eeprom_driver or EEPROMDriver(gpio)
        eeprom_driver.setup()
        gpio_dev: Union[OT3GPIO, RemoteOT3GPIO] = gpio
        usb_messenger: Optional[BinaryMessenger] = None
        if usb_driver:
            usb_messenger = BinaryMessenger(usb_driver)
            usb_messenger.start()
            gpio_dev = RemoteOT3GPIO(usb_messenger)
        return SystemDrivers(
            can_messenger,
            gpio_dev,
            eeprom_driver,
            usb_messenger=usb_messenger,
        )

    @property
    def gear_motor_position(self) -> Optional[float]:
        return self._gear_motor_position.get(NodeId.pipette_left, None)

    def _motor_nodes(self) -> Set[NodeId]:
        """Get a list of the motor controller nodes of all attached and ok devices."""
        return motor_nodes(self._subsystem_manager.targets)

    async def update_firmware(
        self,
        subsystems: Set[SubSystem],
        force: bool = False,
    ) -> AsyncIterator[UpdateStatus]:
        """Updates the firmware on the OT3."""
        async for update in self._subsystem_manager.update_firmware(subsystems, force):
            yield update

    def get_current_settings(
        self, gantry_load: GantryLoad
    ) -> OT3AxisMap[CurrentConfig]:
        return get_current_settings(self._configuration.current_settings, gantry_load)

    async def update_to_default_current_settings(self, gantry_load: GantryLoad) -> None:
        self._current_settings = self.get_current_settings(gantry_load)
        await self.set_default_currents()

    def update_feature_flags(self, feature_flags: HardwareFeatureFlags) -> None:
        """Update the hardware feature flags used by the hardware controller."""
        self._feature_flags = feature_flags

    async def update_motor_status(self) -> None:
        """Retreieve motor and encoder status and position from all present nodes"""
        motor_nodes = self._motor_nodes()
        assert len(motor_nodes)
        response = await get_motor_position(self._messenger, motor_nodes)
        self._handle_motor_status_response(response)

    async def update_motor_estimation(self, axes: Sequence[Axis]) -> None:
        """Update motor position estimation for commanded nodes, and update cache of data."""
        nodes = set([axis_to_node(a) for a in axes])
        response = await update_motor_position_estimation(self._messenger, nodes)
        self._handle_motor_status_response(response)

    @property
    def grip_force_profile(self) -> Optional[GripForceProfile]:
        return self._gripper_force_settings

    @grip_force_profile.setter
    def grip_force_profile(self, profile: Optional[GripForceProfile]) -> None:
        self._gripper_force_settings = profile

    @property
    def motor_run_currents(self) -> OT3AxisMap[float]:
        assert self._current_settings
        run_currents: OT3AxisMap[float] = {}
        for axis, settings in self._current_settings.items():
            run_currents[axis] = settings.run_current
        return run_currents

    @property
    def motor_hold_currents(self) -> OT3AxisMap[float]:
        assert self._current_settings
        hold_currents: OT3AxisMap[float] = {}
        for axis, settings in self._current_settings.items():
            hold_currents[axis] = settings.hold_current
        return hold_currents

    @property
    def gpio_chardev(self) -> Union[OT3GPIO, RemoteOT3GPIO]:
        """Get the GPIO device."""
        return self._gpio_dev

    @property
    def board_revision(self) -> BoardRevision:
        """Get the board revision"""
        return BoardRevision.FLEX_B2

    @property
    def module_controls(self) -> AttachedModulesControl:
        """Get the module controls."""
        if self._module_controls is None:
            raise AttributeError("Module controls not found.")
        return self._module_controls

    @module_controls.setter
    def module_controls(self, module_controls: AttachedModulesControl) -> None:
        """Set the module controls"""
        self._module_controls = module_controls

    def _get_motor_status(
        self, axes: Sequence[Axis]
    ) -> Dict[Axis, Optional[MotorStatus]]:
        return {ax: self._motor_status.get(axis_to_node(ax)) for ax in axes}

    def get_invalid_motor_axes(self, axes: Sequence[Axis]) -> List[Axis]:
        """Get axes that currently do not have the motor-ok flag."""
        return [
            ax
            for ax, status in self._get_motor_status(axes).items()
            if not status or not status.motor_ok
        ]

    def get_invalid_encoder_axes(self, axes: Sequence[Axis]) -> List[Axis]:
        """Get axes that currently do not have the encoder-ok flag."""
        return [
            ax
            for ax, status in self._get_motor_status(axes).items()
            if not status or not status.encoder_ok
        ]

    def check_motor_status(self, axes: Sequence[Axis]) -> bool:
        return len(self.get_invalid_motor_axes(axes)) == 0

    def check_encoder_status(self, axes: Sequence[Axis]) -> bool:
        return len(self.get_invalid_encoder_axes(axes)) == 0

    async def update_position(self) -> OT3AxisMap[float]:
        """Get the current position."""
        return axis_convert(self._position, 0.0)

    async def update_encoder_position(self) -> OT3AxisMap[float]:
        """Get the encoder current position."""
        return axis_convert(self._encoder_position, 0.0)

    def _handle_motor_status_response(
        self, response: NodeMap[MotorPositionStatus], handle_gear_move: bool = False
    ) -> None:
        for axis, pos in response.items():
            if handle_gear_move and axis == NodeId.pipette_left:
                self._gear_motor_position = {axis: pos.motor_position}
            else:
                self._position.update({axis: pos.motor_position})
                self._encoder_position.update({axis: pos.encoder_position})
                # TODO (FPS 6-01-2023): Remove this once the Feature Flag to ignore stall detection is removed.
                # This check will latch the motor status for an axis at "true" if it was ever set to true.
                # To account for the case where a motor axis has its power reset, we also depend on the
                # "encoder_ok" flag staying set (it will only be False if the motor axis has not been
                # homed since a power cycle)
                motor_ok_latch = (
                    (not self._feature_flags.stall_detection_enabled)
                    and (
                        (axis in self._motor_status)
                        and self._motor_status[axis].motor_ok
                    )
                    and self._motor_status[axis].encoder_ok
                )
                self._motor_status.update(
                    {
                        axis: MotorStatus(
                            motor_ok=(pos.motor_ok or motor_ok_latch),
                            encoder_ok=pos.encoder_ok,
                        )
                    }
                )

    def _build_move_node_axis_runner(
        self,
        origin: Dict[Axis, float],
        target: Dict[Axis, float],
        speed: float,
        stop_condition: HWStopCondition,
        nodes_in_moves_only: bool,
    ) -> Tuple[Optional[MoveGroupRunner], bool]:
        if not target:
            return None, False
        move_target = MoveTarget.build(position=target, max_speed=speed)
        try:
            _, movelist = self._move_manager.plan_motion(
                origin=origin, target_list=[move_target]
            )
        except ZeroLengthMoveError as zme:
            log.debug(f"Not moving because move was zero length {str(zme)}")
            return None, False
        moves = movelist[0]
        log.debug(
            f"move: machine coordinates {target} from origin: machine coordinates {origin} at speed: {speed} requires {moves}"
        )

        ordered_nodes = self._motor_nodes()
        if nodes_in_moves_only:
            moving_axes = {
                axis_to_node(ax) for move in moves for ax in move.unit_vector.keys()
            }
            ordered_nodes = ordered_nodes.intersection(moving_axes)

        move_group, _ = create_move_group(
            origin, moves, ordered_nodes, MoveStopCondition[stop_condition.name]
        )
        return (
            MoveGroupRunner(
                move_groups=[move_group],
                ignore_stalls=True
                if not self._feature_flags.stall_detection_enabled
                else False,
            ),
            False,
        )

    def _build_move_gear_axis_runner(
        self,
        possible_q_axis_origin: Optional[float],
        possible_q_axis_target: Optional[float],
        speed: float,
        nodes_in_moves_only: bool,
    ) -> Tuple[Optional[MoveGroupRunner], bool]:
        if possible_q_axis_origin is None or possible_q_axis_target is None:
            return None, True
        tip_motor_move_group = self._build_tip_action_group(
            possible_q_axis_origin, [(possible_q_axis_target, speed)]
        )
        if nodes_in_moves_only:
            ordered_nodes = self._motor_nodes()

            ordered_nodes.intersection({axis_to_node(Axis.Q)})
        return (
            MoveGroupRunner(
                move_groups=[tip_motor_move_group],
                ignore_stalls=True
                if not self._feature_flags.stall_detection_enabled
                else False,
            ),
            True,
        )

    @requires_update
    @requires_estop
    async def move(
        self,
        origin: Dict[Axis, float],
        target: Dict[Axis, float],
        speed: float,
        stop_condition: HWStopCondition = HWStopCondition.none,
        nodes_in_moves_only: bool = True,
    ) -> None:
        """Move to a position.

        Args:
            origin: The starting point of the move
            moves: List of moves.
            stop_condition: The stop condition.
            nodes_in_moves_only: Default is True. If False, also send empty moves to
                                 nodes that are present but not defined in moves.

        .. caution::
            Setting `nodes_in_moves_only` to False will enable *all* present motors in
            the system. DO NOT USE when you want to keep one of the axes disabled.

        Returns:
            None
        """
        possible_q_axis_origin = origin.pop(Axis.Q, None)
        possible_q_axis_target = target.pop(Axis.Q, None)

        maybe_runners = (
            self._build_move_node_axis_runner(
                origin, target, speed, stop_condition, nodes_in_moves_only
            ),
            self._build_move_gear_axis_runner(
                possible_q_axis_origin,
                possible_q_axis_target,
                speed,
                nodes_in_moves_only,
            ),
        )
        log.debug(f"The move groups are {maybe_runners}.")

        gather_moving_nodes = set()
        all_moving_nodes = set()
        for runner, _ in maybe_runners:
            if runner:
                for n in runner.all_nodes():
                    gather_moving_nodes.add(n)
                for n in runner.all_moving_nodes():
                    all_moving_nodes.add(n)

        pipettes_moving = moving_pipettes_in_move_group(
            gather_moving_nodes, all_moving_nodes
        )

        async def _runner_coroutine(
            runner: MoveGroupRunner, is_gear_move: bool
        ) -> Tuple[Dict[NodeId, MotorPositionStatus], bool]:
            positions = await runner.run(can_messenger=self._messenger)
            return positions, is_gear_move

        coros = [
            _runner_coroutine(runner, is_gear_move)
            for runner, is_gear_move in maybe_runners
            if runner
        ]
        checked_moving_pipettes = self._pipettes_to_monitor_pressure(pipettes_moving)
        async with self._monitor_overpressure(checked_moving_pipettes):
            all_positions = await asyncio.gather(*coros)

        for positions, handle_gear_move in all_positions:
            self._handle_motor_status_response(positions, handle_gear_move)

    def _get_axis_home_distance(self, axis: Axis) -> float:
        if self.check_motor_status([axis]):
            return -1 * (
                self._position[axis_to_node(axis)] + LIMIT_SWITCH_OVERTRAVEL_DISTANCE
            )
        else:
            return -1 * self.axis_bounds[axis][1] - self.axis_bounds[axis][0]

    def _build_axes_home_groups(
        self, axes: Sequence[Axis], speed_settings: Dict[OT3AxisKind, float]
    ) -> List[MoveGroup]:
        present_axes = [ax for ax in axes if self.axis_is_present(ax)]
        if not present_axes:
            return []
        else:
            distances = {ax: self._get_axis_home_distance(ax) for ax in present_axes}
            velocities = {
                ax: -1 * speed_settings[Axis.to_kind(ax)] for ax in present_axes
            }
            return create_home_groups(distances, velocities)

    def _build_home_pipettes_runner(
        self,
        axes: Sequence[Axis],
        gantry_load: GantryLoad,
    ) -> Optional[MoveGroupRunner]:
        pipette_axes = [ax for ax in axes if ax in Axis.pipette_axes()]
        if not pipette_axes:
            return None

        speed_settings = self._configuration.motion_settings.max_speed_discontinuity[
            gantry_load
        ]
        move_groups: List[MoveGroup] = self._build_axes_home_groups(
            pipette_axes, speed_settings
        )
        return MoveGroupRunner(move_groups=move_groups)

    def _build_home_gantry_z_runner(
        self,
        axes: Sequence[Axis],
        gantry_load: GantryLoad,
    ) -> Optional[MoveGroupRunner]:
        gantry_axes = [ax for ax in axes if ax in Axis.gantry_axes()]
        if not gantry_axes:
            return None

        speed_settings = self._configuration.motion_settings.max_speed_discontinuity[
            gantry_load
        ]

        # first home all the present mount axes
        z_axes = list(filter(lambda ax: ax in Axis.ot3_mount_axes(), gantry_axes))
        z_groups = self._build_axes_home_groups(z_axes, speed_settings)

        # home X axis before Y axis, to avoid collision with thermo-cycler lid
        # that could be in the back-left corner
        x_groups = (
            self._build_axes_home_groups([Axis.X], speed_settings)
            if Axis.X in gantry_axes
            else []
        )
        y_groups = (
            self._build_axes_home_groups([Axis.Y], speed_settings)
            if Axis.Y in gantry_axes
            else []
        )

        move_groups = [*z_groups, *x_groups, *y_groups]
        if move_groups:
            return MoveGroupRunner(move_groups=move_groups)
        return None

    @requires_update
    @requires_estop
    async def home(
        self, axes: Sequence[Axis], gantry_load: GantryLoad
    ) -> OT3AxisMap[float]:
        """Home each axis passed in, and reset the positions to 0.

        Args:
            axes: List[Axis]

        Returns:
            A dictionary containing the new positions of each axis
        """
        checked_axes = [axis for axis in axes if self.axis_is_present(axis)]
        assert Axis.G not in checked_axes, "Please home G axis using gripper_home_jaw()"
        if not checked_axes:
            return {}

        maybe_runners = (
            self._build_home_gantry_z_runner(checked_axes, gantry_load),
            self._build_home_pipettes_runner(checked_axes, gantry_load),
        )
        coros = [
            runner.run(can_messenger=self._messenger)
            for runner in maybe_runners
            if runner
        ]
        moving_pipettes = [
            axis_to_node(ax) for ax in checked_axes if ax in Axis.pipette_axes()
        ]
        checked_moving_pipettes = self._pipettes_to_monitor_pressure(moving_pipettes)
        async with self._monitor_overpressure(checked_moving_pipettes):
            positions = await asyncio.gather(*coros)
        # TODO(CM): default gear motor homing routine to have some acceleration
        if Axis.Q in checked_axes:
            await self.home_tip_motors(
                distance=self.axis_bounds[Axis.Q][1] - self.axis_bounds[Axis.Q][0],
                velocity=self._configuration.motion_settings.max_speed_discontinuity.high_throughput[
                    Axis.to_kind(Axis.Q)
                ],
            )

        for position in positions:
            self._handle_motor_status_response(position)
        return axis_convert(self._position, 0.0)

    def _pipettes_to_monitor_pressure(self, pipettes: List[NodeId]) -> List[NodeId]:
        return [pip for pip in pipettes if self._pressure_sensor_available[pip]]

    def _filter_move_group(self, move_group: MoveGroup) -> MoveGroup:
        new_group: MoveGroup = []
        for step in move_group:
            new_group.append(
                {
                    node: axis_step
                    for node, axis_step in step.items()
                    if node in self._motor_nodes()
                }
            )
        return new_group

    async def home_tip_motors(
        self,
        distance: float,
        velocity: float,
        back_off: bool = True,
    ) -> None:
        move_group = create_tip_motor_home_group(distance, velocity, back_off)

        runner = MoveGroupRunner(
            move_groups=[move_group],
            ignore_stalls=True
            if not self._feature_flags.stall_detection_enabled
            else False,
        )
        try:
            positions = await runner.run(can_messenger=self._messenger)
            if NodeId.pipette_left in positions:
                self._gear_motor_position = {
                    NodeId.pipette_left: positions[NodeId.pipette_left].motor_position
                }
            else:
                log.debug("no position returned from NodeId.pipette_left")
                self._gear_motor_position = {}
        except Exception as e:
            log.error("Clearing tip motor position due to failed movement")
            self._gear_motor_position = {}
            raise e

    def _build_tip_action_group(
        self, origin: float, targets: List[Tuple[float, float]]
    ) -> MoveGroup:
        move_targets = [
            MoveTarget.build({Axis.Q: target_pos}, speed)
            for target_pos, speed in targets
        ]
        _, moves = self._move_manager.plan_motion(
            origin={Axis.Q: origin}, target_list=move_targets
        )

        return create_tip_action_group(moves[0], [NodeId.pipette_left], "clamp")

    async def tip_action(
        self, origin: float, targets: List[Tuple[float, float]]
    ) -> None:
        move_group = self._build_tip_action_group(origin, targets)
        runner = MoveGroupRunner(
            move_groups=[move_group],
            ignore_stalls=True
            if not self._feature_flags.stall_detection_enabled
            else False,
        )
        try:
            positions = await runner.run(can_messenger=self._messenger)
            if NodeId.pipette_left in positions:
                self._gear_motor_position = {
                    NodeId.pipette_left: positions[NodeId.pipette_left].motor_position
                }
            else:
                log.debug("no position returned from NodeId.pipette_left")
                self._gear_motor_position = {}
        except Exception as e:
            log.error("Clearing tip motor position due to failed movement")
            self._gear_motor_position = {}
            raise e

    @requires_update
    @requires_estop
    async def gripper_grip_jaw(
        self,
        duty_cycle: float,
        expected_displacement: float,  # not used on real hardware
        stop_condition: HWStopCondition = HWStopCondition.none,
        stay_engaged: bool = True,
    ) -> None:
        move_group = create_gripper_jaw_grip_group(
            duty_cycle, MoveStopCondition[stop_condition.name], stay_engaged
        )
        runner = MoveGroupRunner(move_groups=[move_group])
        positions = await runner.run(can_messenger=self._messenger)
        self._handle_motor_status_response(positions)

    @requires_update
    @requires_estop
    async def gripper_hold_jaw(
        self,
        encoder_position_um: int,
    ) -> None:
        move_group = create_gripper_jaw_hold_group(encoder_position_um)
        runner = MoveGroupRunner(move_groups=[move_group])
        positions = await runner.run(can_messenger=self._messenger)
        self._handle_motor_status_response(positions)

    @requires_update
    @requires_estop
    async def gripper_home_jaw(self, duty_cycle: float) -> None:
        move_group = create_gripper_jaw_home_group(duty_cycle)
        runner = MoveGroupRunner(move_groups=[move_group])
        positions = await runner.run(can_messenger=self._messenger)
        self._handle_motor_status_response(positions)

    async def get_jaw_state(self) -> GripperJawState:
        res = await get_gripper_jaw_state(self._messenger)
        return gripper_jaw_state_from_fw(res)

    @staticmethod
    def _lookup_serial_key(pipette_name: FirmwarePipetteName) -> str:
        lookup_name = {
            FirmwarePipetteName.p1000_single: "P1KS",
            FirmwarePipetteName.p1000_multi: "P1KM",
            FirmwarePipetteName.p1000_multi_em: "P1KP",
            FirmwarePipetteName.p50_single: "P50S",
            FirmwarePipetteName.p50_multi: "P50M",
            FirmwarePipetteName.p1000_96: "P1KH",
            FirmwarePipetteName.p50_96: "P50H",
            FirmwarePipetteName.p200_96: "P2HH",
        }
        return lookup_name[pipette_name]

    @staticmethod
    def _combine_serial_number(pipette_info: ohc_tool_types.PipetteInformation) -> str:
        serialized_name = OT3Controller._lookup_serial_key(pipette_info.name)
        version = pipette_load_name.version_from_string(pipette_info.model)
        return f"{serialized_name}V{version.major}{version.minor}{pipette_info.serial}"

    @staticmethod
    def _build_attached_pip(
        attached: ohc_tool_types.PipetteInformation, mount: OT3Mount
    ) -> AttachedPipette:
        if attached.name == FirmwarePipetteName.unknown:
            raise InvalidPipetteName(name=attached.name_int, mount=mount.name)
        try:
            # TODO (lc 12-8-2022) We should return model as an int rather than
            # a string.
            # TODO (lc 12-6-2022) We should also provide the full serial number
            # for PipetteInformation.serial so we don't have to use
            # helper methods to convert the serial back to what was flashed
            # on the eeprom.
            converted_name = pipette_load_name.convert_pipette_name(
                cast(PipetteName, attached.name.name), attached.model
            )
            return {
                "config": load_pipette_data.load_definition(
                    converted_name.pipette_type,
                    converted_name.pipette_channels,
                    converted_name.pipette_version,
                    converted_name.oem_type,
                ),
                "id": OT3Controller._combine_serial_number(attached),
            }
        except KeyError:
            raise InvalidPipetteModel(
                name=attached.name.name, model=attached.model, mount=mount.name
            )

    @staticmethod
    def _build_attached_gripper(
        attached: ohc_tool_types.GripperInformation,
    ) -> AttachedGripper:
        model = gripper_config.info_num_to_model(attached.model)
        serial = attached.serial
        return {
            "config": gripper_config.load(model),
            "id": f"GRPV{attached.model.replace('.', '')}{serial}",
        }

    @staticmethod
    def _generate_attached_instrs(
        attached: ohc_tool_types.ToolSummary,
    ) -> Iterator[Tuple[OT3Mount, OT3AttachedInstruments]]:
        if attached.left:
            yield (
                OT3Mount.LEFT,
                OT3Controller._build_attached_pip(attached.left, OT3Mount.LEFT),
            )
        if attached.right:
            yield (
                OT3Mount.RIGHT,
                OT3Controller._build_attached_pip(attached.right, OT3Mount.RIGHT),
            )
        if attached.gripper:
            yield (
                OT3Mount.GRIPPER,
                OT3Controller._build_attached_gripper(attached.gripper),
            )

    async def get_attached_instruments(
        self, expected: Mapping[OT3Mount, PipetteName]
    ) -> Dict[OT3Mount, OT3AttachedInstruments]:
        """Get attached instruments.

        Args:
            expected: Which mounts are expected.

        Returns:
            A map of mount to instrument name.
        """
        return dict(
            OT3Controller._generate_attached_instrs(self._subsystem_manager.tools)
        )

    async def get_limit_switches(self) -> OT3AxisMap[bool]:
        """Get the state of the gantry's limit switches on each axis."""
        motor_nodes = self._motor_nodes()
        assert motor_nodes, "No nodes available to read limit switch status from"
        res = await get_limit_switches(self._messenger, motor_nodes)
        return {node_to_axis(node): bool(val) for node, val in res.items()}

    @staticmethod
    def _tip_motor_nodes(axis_current_keys: KeysView[Axis]) -> List[NodeId]:
        return [axis_to_node(Axis.Q)] if Axis.Q in axis_current_keys else []

    async def set_default_currents(self) -> None:
        """Set both run and hold currents from robot config to each node."""
        assert self._current_settings, "Invalid current settings"
        await set_currents(
            self._messenger,
            self._axis_map_to_present_nodes(
                {k: v.as_tuple() for k, v in self._current_settings.items()}
            ),
            use_tip_motor_message_for=self._tip_motor_nodes(
                self._current_settings.keys()
            ),
        )

    @requires_update
    async def set_active_current(self, axis_currents: OT3AxisMap[float]) -> None:
        """Set the active current.

        Args:
            axis_currents: Axes' currents

        Returns:
            None
        """
        assert self._current_settings, "Invalid current settings"
        await set_run_current(
            self._messenger,
            self._axis_map_to_present_nodes(axis_currents),
            use_tip_motor_message_for=self._tip_motor_nodes(axis_currents.keys()),
        )
        for axis, current in axis_currents.items():
            self._current_settings[axis].run_current = current

    @requires_update
    async def set_hold_current(self, axis_currents: OT3AxisMap[float]) -> None:
        """Set the hold current for motor.

        Args:
            axis_currents: Axes' currents

        Returns:
            None
        """
        assert self._current_settings, "Invalid current settings"
        await set_hold_current(
            self._messenger,
            self._axis_map_to_present_nodes(axis_currents),
            use_tip_motor_message_for=self._tip_motor_nodes(axis_currents.keys()),
        )
        for axis, current in axis_currents.items():
            self._current_settings[axis].hold_current = current

    @asynccontextmanager
    async def motor_current(
        self,
        run_currents: Optional[OT3AxisMap[float]] = None,
        hold_currents: Optional[OT3AxisMap[float]] = None,
    ) -> AsyncIterator[None]:
        """Update and restore current."""
        assert self._current_settings
        old_settings = deepcopy(self._current_settings)
        if run_currents:
            await self.set_active_current(run_currents)
        if hold_currents:
            await self.set_hold_current(hold_currents)
        try:
            yield
        finally:
            if run_currents:
                await self.set_active_current(
                    {ax: old_settings[ax].run_current for ax in run_currents.keys()}
                )
            if hold_currents:
                await self.set_hold_current(
                    {ax: old_settings[ax].hold_current for ax in hold_currents.keys()}
                )
            if not run_currents and not hold_currents:
                self._current_settings = old_settings
                await self.set_default_currents()

    @asynccontextmanager
    async def restore_z_r_run_current(self) -> AsyncIterator[None]:
        """
        Temporarily restore the active current ONLY when homing or
        retracting the Z_R axis while the 96-channel is attached.
        """
        assert self._current_settings
        high_throughput_settings = deepcopy(self._current_settings)
        conf = self.get_current_settings(GantryLoad.LOW_THROUGHPUT)[Axis.Z_R]
        # outside of homing and retracting, Z_R run current should
        # be reduced to its hold current
        await self.set_active_current({Axis.Z_R: conf.run_current})
        try:
            yield
        finally:
            await self.set_active_current(
                {Axis.Z_R: high_throughput_settings[Axis.Z_R].run_current}
            )

    @asynccontextmanager
    async def increase_z_l_hold_current(self) -> AsyncIterator[None]:
        """
        Temporarily increase the hold current when engaging the Z_L axis
        while the 96-channel is attached
        """
        assert self._current_settings
        high_throughput_settings = deepcopy(self._current_settings)
        await self.set_hold_current(
            {Axis.Z_L: high_throughput_settings[Axis.Z_L].run_current}
        )
        try:
            yield
        finally:
            await self.set_hold_current(
                {Axis.Z_L: high_throughput_settings[Axis.Z_L].hold_current}
            )

    @staticmethod
    def _build_event_watcher() -> aionotify.Watcher:
        watcher = aionotify.Watcher()
        watcher.watch(
            alias="modules",
            path="/dev",
            flags=(
                aionotify.Flags.CREATE
                | aionotify.Flags.DELETE
                | aionotify.Flags.MOVED_FROM
                | aionotify.Flags.MOVED_TO
            ),
        )
        return watcher

    async def _handle_watch_event(self) -> None:
        try:
            event = await self._event_watcher.get_event()
        except asyncio.IncompleteReadError:
            log.debug("incomplete read error when quitting watcher")
            return
        if event is not None:
            flags = aionotify.Flags.parse(event.flags)
            log.debug(f"aionotify: {flags} {event.name}")
            if "ot_module" in event.name:
                event_name = event.name
                event_description = AionotifyEvent.build(event_name, flags)
                await self.module_controls.handle_module_appearance(event_description)

    async def watch(self, loop: asyncio.AbstractEventLoop) -> None:
        can_watch = aionotify is not None
        if can_watch:
            await self._event_watcher.setup(loop)

        while can_watch and (not self._event_watcher.closed):
            await self._handle_watch_event()

    @property
    def axis_bounds(self) -> OT3AxisMap[Tuple[float, float]]:
        """Get the axis bounds."""
        # TODO (AL, 2021-11-18): The bounds need to be defined
        return {
            Axis.Z_L: (0, 300),
            Axis.Z_R: (0, 300),
            Axis.P_L: (0, 200),
            Axis.P_R: (0, 200),
            Axis.X: (0, 550),
            Axis.Y: (0, 550),
            Axis.Z_G: (0, 300),
            Axis.Q: (0, 200),
        }

    def engaged_axes(self) -> OT3AxisMap[bool]:
        """Get engaged axes."""
        return self._engaged_axes

    async def update_engaged_axes(self) -> None:
        """Update engaged axes."""
        motor_nodes = self._motor_nodes()
        results = await get_motor_enabled(self._messenger, motor_nodes)
        for node, status in results.items():
            self._engaged_axes[node_to_axis(node)] = status

    async def is_motor_engaged(self, axis: Axis) -> bool:
        node = axis_to_node(axis)
        result = await get_motor_enabled(self._messenger, {node})
        try:
            engaged = result[node]
        except KeyError as ke:
            raise CommunicationError(
                message=f"No response from {node.name} for motor engagement query",
                detail={"node": node.name},
                wrapping=[PythonException(ke)],
            ) from ke
        self._engaged_axes.update({axis: engaged})
        return engaged

    async def disengage_axes(self, axes: List[Axis]) -> None:
        """Disengage axes."""
        if Axis.Q in axes:
            await set_disable_tip_motor(self._messenger, {axis_to_node(Axis.Q)})
            self._engaged_axes[Axis.Q] = False
            axes = [ax for ax in axes if ax is not Axis.Q]

        if len(axes) > 0:
            await set_disable_motor(self._messenger, {axis_to_node(ax) for ax in axes})
        for ax in axes:
            self._engaged_axes[ax] = False

    async def engage_axes(self, axes: List[Axis]) -> None:
        """Engage axes."""
        if Axis.Q in axes:
            await set_enable_tip_motor(self._messenger, {axis_to_node(Axis.Q)})
            self._engaged_axes[Axis.Q] = True
            axes = [ax for ax in axes if ax is not Axis.Q]

        if len(axes) > 0:
            await set_enable_motor(self._messenger, {axis_to_node(ax) for ax in axes})
        for ax in axes:
            self._engaged_axes[ax] = True

    @requires_update
    async def set_lights(self, button: Optional[bool], rails: Optional[bool]) -> None:
        """Set the light states."""
        if rails is not None:
            await set_deck_light(1 if rails else 0, self._usb_messenger)

    @requires_update
    async def get_lights(self) -> Dict[str, bool]:
        """Get the light state."""
        return {
            "rails": await get_deck_light_state(self._usb_messenger),
            "button": False,
        }

    def pause(self) -> None:
        """Pause the controller activity."""
        return None

    def resume(self) -> None:
        """Resume the controller activity."""
        return None

    async def halt(self) -> None:
        """Halt the motors."""
        error = await self._messenger.ensure_send(
            NodeId.broadcast, StopRequest(payload=EmptyPayload())
        )
        if error != ErrorCode.ok:
            log.warning(f"Halt stop request failed: {error}")

    async def probe(self, axis: Axis, distance: float) -> OT3AxisMap[float]:
        """Probe."""
        return {}

    async def clean_up(self) -> None:
        """Clean up."""
        try:
            loop = asyncio.get_event_loop()
        except RuntimeError:
            return

        if hasattr(self, "_event_watcher"):
            if (
                loop.is_running()
                and self._event_watcher
                and not self._event_watcher.closed
            ):
                self._event_watcher.close()

        messenger = getattr(self, "_messenger", None)
        if messenger:
            await messenger.stop()

        usb_messenger = getattr(self, "_usb_messenger", None)
        if usb_messenger:
            await usb_messenger.stop()

        return None

    @staticmethod
    def _get_home_position() -> Dict[NodeId, float]:
        return {
            NodeId.head_l: 0,
            NodeId.head_r: 0,
            NodeId.gantry_x: 0,
            NodeId.gantry_y: 0,
            NodeId.pipette_left: 0,
            NodeId.pipette_right: 0,
            NodeId.gripper_z: 0,
            NodeId.gripper_g: 0,
        }

    @staticmethod
    def home_position() -> OT3AxisMap[float]:
        return {
            node_to_axis(k): v for k, v in OT3Controller._get_home_position().items()
        }

    async def probe_network(self, timeout: float = 5.0) -> None:
        """Update the list of nodes present on the network.

        The stored result is used to make sure that move commands include entries
        for all present axes, so none incorrectly move before the others are ready.
        """
        await self._subsystem_manager.refresh()

    def axis_is_present(self, axis: Axis) -> bool:
        try:
            return axis_to_node(axis) in self._motor_nodes()
        except KeyError:
            # Currently unhandled axis
            return False

    def _axis_map_to_present_nodes(
        self, to_xform: OT3AxisMap[MapPayload]
    ) -> NodeMap[MapPayload]:
        by_node = {axis_to_node(k): v for k, v in to_xform.items()}
        return {k: v for k, v in by_node.items() if k in self._motor_nodes()}

    @asynccontextmanager
    async def _monitor_overpressure(self, mounts: List[NodeId]) -> AsyncIterator[None]:
        msg = "The pressure sensor on the {} mount has exceeded operational limits."
        if self._feature_flags.overpressure_detection_enabled and mounts:
            tools_with_id = map_pipette_type_to_sensor_id(
                mounts, self._subsystem_manager.device_info
            )
            # FIXME we should switch the sensor type based on the channel
            # used when partial tip pick up is implemented.
            provided_context_manager = await check_overpressure(
                self._messenger, tools_with_id
            )
            errors: asyncio.Queue[Tuple[NodeId, ErrorCode]] = asyncio.Queue()

            async with provided_context_manager() as errors:
                try:
                    yield
                finally:

                    def _pop_queue() -> Optional[Tuple[NodeId, ErrorCode]]:
                        try:
                            return errors.get_nowait()
                        except asyncio.QueueEmpty:
                            return None

                    q_msg = _pop_queue()
                    if q_msg:
                        mount = Axis.to_ot3_mount(node_to_axis(q_msg[0]))
                        raise PipetteOverpressureError(
                            message=msg.format(str(mount)),
                            detail={"mount": str(mount)},
                        )
        else:
            yield

    async def liquid_probe(
        self,
        mount: OT3Mount,
        max_p_distance: float,
        mount_speed: float,
        plunger_speed: float,
        threshold_pascals: float,
        plunger_impulse_time: float,
        num_baseline_reads: int,
        z_offset_for_plunger_prep: float,
        probe: InstrumentProbeType = InstrumentProbeType.PRIMARY,
        force_both_sensors: bool = False,
        response_queue: Optional[PipetteSensorResponseQueue] = None,
    ) -> float:
        head_node = axis_to_node(Axis.by_mount(mount))
        tool = sensor_node_for_pipette(OT3Mount(mount.value))
        if tool not in self._pipettes_to_monitor_pressure([tool]):
            raise UnsupportedHardwareCommand(
                "Liquid Presence Detection not available on this pipette."
            )

        if response_queue is None:
            response_capture: Optional[
                Callable[[Dict[SensorId, List[SensorDataType]]], None]
            ] = None
        else:

            def response_capture(data: Dict[SensorId, List[SensorDataType]]) -> None:
                response_queue.put_nowait(
                    {
                        PipetteSensorId(sensor_id.value): [
                            PipetteSensorData(
                                sensor_type=PipetteSensorType(packet.sensor_type.value),
                                _as_int=packet.to_int,
                                _as_float=packet.to_float(),
                            )
                            for packet in packets
                        ]
                        for sensor_id, packets in data.items()
                    }
                )

        positions = await liquid_probe(
            messenger=self._messenger,
            tool=tool,
            head_node=head_node,
            max_p_distance=max_p_distance,
            plunger_speed=plunger_speed,
            mount_speed=mount_speed,
            threshold_pascals=threshold_pascals,
            plunger_impulse_time=plunger_impulse_time,
            num_baseline_reads=num_baseline_reads,
            z_offset_for_plunger_prep=z_offset_for_plunger_prep,
            sensor_id=sensor_id_for_instrument(probe),
            force_both_sensors=force_both_sensors,
            emplace_data=response_capture,
        )
        for node, point in positions.items():
            self._position.update({node: point.motor_position})
            self._encoder_position.update({node: point.encoder_position})
        if (
            head_node not in positions
            or positions[head_node].move_ack
            == MoveCompleteAck.complete_without_condition
        ):
            raise PipetteLiquidNotFoundError(
                "Liquid not found during probe.",
                {
                    str(node_to_axis(node)): str(point.motor_position)
                    for node, point in positions.items()
                },
            )
        return self._position[axis_to_node(Axis.by_mount(mount))]

    async def capacitive_probe(
        self,
        mount: OT3Mount,
        moving: Axis,
        distance_mm: float,
        speed_mm_per_s: float,
        sensor_threshold_pf: float,
        probe: InstrumentProbeType = InstrumentProbeType.PRIMARY,
    ) -> bool:
        status = await capacitive_probe(
            messenger=self._messenger,
            tool=sensor_node_for_mount(mount),
            mover=axis_to_node(moving),
            distance=distance_mm,
            mount_speed=speed_mm_per_s,
            sensor_id=sensor_id_for_instrument(probe),
            relative_threshold_pf=sensor_threshold_pf,
        )

        self._position[axis_to_node(moving)] = status.motor_position
        return status.move_ack == MoveCompleteAck.stopped_by_condition

    async def capacitive_pass(
        self,
        mount: OT3Mount,
        moving: Axis,
        distance_mm: float,
        speed_mm_per_s: float,
        probe: InstrumentProbeType,
    ) -> List[float]:
        data = await capacitive_pass(
            self._messenger,
            sensor_node_for_mount(mount),
            axis_to_node(moving),
            distance_mm,
            speed_mm_per_s,
            sensor_id_for_instrument(probe),
        )
        self._position[axis_to_node(moving)] += distance_mm
        return data

    async def release_estop(self) -> None:
        if self._gpio_dev is None:
            log.error("no gpio control available")
            raise IOError("no gpio control")
        elif isinstance(self._gpio_dev, RemoteOT3GPIO):
            await self._gpio_dev.deactivate_estop()
        else:
            self._gpio_dev.deactivate_estop()

    async def engage_estop(self) -> None:
        if self._gpio_dev is None:
            log.error("no gpio control available")
            raise IOError("no gpio control")
        elif isinstance(self._gpio_dev, RemoteOT3GPIO):
            await self._gpio_dev.activate_estop()
        else:
            self._gpio_dev.activate_estop()

    async def release_sync(self) -> None:
        if self._gpio_dev is None:
            log.error("no gpio control available")
            raise IOError("no gpio control")
        elif isinstance(self._gpio_dev, RemoteOT3GPIO):
            await self._gpio_dev.deactivate_nsync_out()
        else:
            self._gpio_dev.deactivate_nsync_out()

    async def engage_sync(self) -> None:
        if self._gpio_dev is None:
            log.error("no gpio control available")
            raise IOError("no gpio control")
        elif isinstance(self._gpio_dev, RemoteOT3GPIO):
            await self._gpio_dev.activate_nsync_out()
        else:
            self._gpio_dev.activate_nsync_out()

    async def door_state(self) -> DoorState:
        door_open = await get_door_state(self._usb_messenger)
        return DoorState.OPEN if door_open else DoorState.CLOSED

    def add_door_state_listener(self, callback: Callable[[DoorState], None]) -> None:
        def _door_listener(msg: BinaryMessageDefinition) -> None:
            door_state = (
                DoorState.OPEN
                if cast(DoorSwitchStateInfo, msg).door_open.value
                else DoorState.CLOSED
            )
            callback(door_state)

        if self._usb_messenger is not None:
            self._usb_messenger.add_listener(
                _door_listener,
                lambda message_id: bool(
                    message_id == BinaryMessageId.door_switch_state_info
                ),
            )

    async def build_estop_detector(self) -> bool:
        """Must be called to set up the estop detector & state machine."""
        if self._drivers.usb_messenger is None:
            return False
        self._estop_detector = await EstopDetector.build(
            usb_messenger=self._drivers.usb_messenger
        )
        self._estop_state_machine.subscribe_to_detector(self._estop_detector)
        return True

    @property
    def tip_presence_manager(self) -> TipPresenceManager:
        return self._tip_presence_manager

    async def update_tip_detector(self, mount: OT3Mount, sensor_count: int) -> None:
        """Build indiviudal tip detector for a mount."""
        await self.teardown_tip_detector(mount)
        await self._tip_presence_manager.build_detector(mount, sensor_count)

    async def teardown_tip_detector(self, mount: OT3Mount) -> None:
        await self._tip_presence_manager.clear_detector(mount)

    async def get_tip_status(
        self,
        mount: OT3Mount,
        follow_singular_sensor: Optional[InstrumentProbeType] = None,
    ) -> TipStateType:
        return await self.tip_presence_manager.get_tip_status(
            mount, follow_singular_sensor
        )

    def current_tip_state(self, mount: OT3Mount) -> Optional[bool]:
        return self.tip_presence_manager.current_tip_state(mount)

    async def set_status_bar_state(self, state: StatusBarState) -> None:
        await self._status_bar_controller.set_status_bar_state(state)

    async def set_status_bar_enabled(self, enabled: bool) -> None:
        await self._status_bar_controller.set_enabled(enabled)

    def get_status_bar_state(self) -> StatusBarState:
        return self._status_bar_controller.get_current_state()

    @property
    def estop_status(self) -> EstopOverallStatus:
        return EstopOverallStatus(
            state=self._estop_state_machine.state,
            left_physical_state=self._estop_state_machine.get_physical_status(
                EstopAttachLocation.LEFT
            ),
            right_physical_state=self._estop_state_machine.get_physical_status(
                EstopAttachLocation.RIGHT
            ),
        )

    def estop_acknowledge_and_clear(self) -> EstopOverallStatus:
        """Attempt to acknowledge an Estop event and clear the status.

        Returns the estop status after clearing the status."""
        self._estop_state_machine.acknowledge_and_clear()
        return self.estop_status

    def get_estop_state(self) -> EstopState:
        return self._estop_state_machine.state

    def add_estop_callback(self, cb: HardwareEventHandler) -> HardwareEventUnsubscriber:
        return self._estop_state_machine.add_listener(cb)

    def check_gripper_position_within_bounds(
        self,
        expected_grip_width: float,
        grip_width_uncertainty_wider: float,
        grip_width_uncertainty_narrower: float,
        jaw_width: float,
        max_allowed_grip_error: float,
        hard_limit_lower: float,
        hard_limit_upper: float,
    ) -> None:
        """
        Check if the gripper is at the expected location.

        While this doesn't seem like it belongs here, it needs to act differently
        when we're simulating, so it does.
        """
        expected_gripper_position_min = (
            expected_grip_width - grip_width_uncertainty_narrower
        )
        expected_gripper_position_max = (
            expected_grip_width + grip_width_uncertainty_wider
        )
        current_gripper_position = jaw_width
        if isclose(current_gripper_position, hard_limit_lower):
            raise FailedGripperPickupError(
                message="Failed to grip: jaws all the way closed",
                details={
                    "failure-type": "jaws-all-the-way-closed",
                    "actual-jaw-width": current_gripper_position,
                },
            )
        if isclose(current_gripper_position, hard_limit_upper):
            raise FailedGripperPickupError(
                message="Failed to grip: jaws all the way open",
                details={
                    "failure-type": "jaws-all-the-way-open",
                    "actual-jaw-width": current_gripper_position,
                },
            )
        if (
            current_gripper_position - expected_gripper_position_min
            < -max_allowed_grip_error
        ):
            raise FailedGripperPickupError(
                message="Failed to grip: jaws closed too far",
                details={
                    "failure-type": "jaws-more-closed-than-expected",
                    "lower-bound-labware-width": expected_grip_width
                    - grip_width_uncertainty_narrower,
                    "actual-jaw-width": current_gripper_position,
                },
            )
        if (
            current_gripper_position - expected_gripper_position_max
            > max_allowed_grip_error
        ):
            raise FailedGripperPickupError(
                message="Failed to grip: jaws could not close far enough",
                details={
                    "failure-type": "jaws-more-open-than-expected",
                    "upper-bound-labware-width": expected_grip_width
                    - grip_width_uncertainty_narrower,
                    "actual-jaw-width": current_gripper_position,
                },
            )

    async def set_hepa_fan_state(self, fan_on: bool, duty_cycle: int) -> bool:
        return await set_hepa_fan_state_fw(self._messenger, fan_on, duty_cycle)

    async def get_hepa_fan_state(self) -> Optional[HepaFanState]:
        res = await get_hepa_fan_state_fw(self._messenger)
        return (
            HepaFanState(
                fan_on=res.fan_on,
                duty_cycle=res.duty_cycle,
            )
            if res
            else None
        )

    async def set_hepa_uv_state(self, light_on: bool, uv_duration_s: int) -> bool:
        return await set_hepa_uv_state_fw(self._messenger, light_on, uv_duration_s)

    async def get_hepa_uv_state(self) -> Optional[HepaUVState]:
        res = await get_hepa_uv_state_fw(self._messenger)
        return (
            HepaUVState(
                light_on=res.uv_light_on,
                uv_duration_s=res.uv_duration_s,
                remaining_time_s=res.remaining_time_s,
            )
            if res
            else None
        )

    def _update_tip_state(self, mount: OT3Mount, status: bool) -> None:
        """This is something we only use in the simulator.
        It is required so that PE simulations using ot3api don't break."""
        pass

    async def increase_evo_disp_count(self, mount: OT3Mount) -> None:
        """Tell a pipette to increase it's evo-tip-dispense-count in eeprom."""
        await send_evo_dispense_count_increase(
            self._messenger, sensor_node_for_pipette(OT3Mount(mount.value))
        )

    async def _read_env_sensor(
        self, mount: OT3Mount, primary: bool
    ) -> Optional[EnvironmentSensorDataType]:
        """Read and return the current sensor information."""
        sensor = EnvironmentSensor.build(
            sensor_id=SensorId.S0 if primary else SensorId.S1,
            node_id=sensor_node_for_mount(mount),
        )
        s_driver = SensorDriver()
        sensor_data = await s_driver.read(
            can_messenger=self._messenger,
            sensor=sensor,
            offset=False,
        )
        assert sensor_data is None or isinstance(sensor_data, EnvironmentSensorDataType)
        return sensor_data

    async def read_env_temp_sensor(
        self, mount: OT3Mount, primary: bool
    ) -> Optional[float]:
        """Read and return the current sensor information."""
        s_data = await self._read_env_sensor(mount, primary)
        if s_data is None or s_data.temperature is None:
            return None
        return s_data.temperature.to_float()

    async def read_env_hum_sensor(
        self, mount: OT3Mount, primary: bool
    ) -> Optional[float]:
        """Read and return the current sensor information."""
        s_data = await self._read_env_sensor(mount, primary)
        if s_data is None or s_data.humidity is None:
            return None
        return s_data.humidity.to_float()

    async def read_pressure_sensor(
        self, mount: OT3Mount, primary: bool
    ) -> Optional[float]:
        """Read and return the current sensor information."""
        sensor = PressureSensor.build(
            sensor_id=SensorId.S0 if primary else SensorId.S1,
            node_id=sensor_node_for_mount(mount),
        )
        s_driver = SensorDriver()
        sensor_data = await s_driver.read(
            can_messenger=self._messenger,
            sensor=sensor,
            offset=False,
        )
        assert sensor_data is None or isinstance(sensor_data, SensorDataType)
        return sensor_data.to_float() if sensor_data else None

    async def read_capacitive_sensor(
        self, mount: OT3Mount, primary: bool
    ) -> Optional[float]:
        """Read and return the current sensor information."""
        sensor = CapacitiveSensor.build(
            sensor_id=SensorId.S0 if primary else SensorId.S1,
            node_id=sensor_node_for_mount(mount),
        )
        s_driver = SensorDriver()
        sensor_data = await s_driver.read(
            can_messenger=self._messenger,
            sensor=sensor,
            offset=False,
        )
        assert sensor_data is None or isinstance(sensor_data, SensorDataType)
        return sensor_data.to_float() if sensor_data else None



"""Protocol engine types to do with engine execution."""
from enum import Enum


# todo(mm, 2024-06-24): This monolithic status field is getting to be a bit much.
# We should consider splitting this up into multiple fields.
class EngineStatus(str, Enum):
    """Current execution status of a ProtocolEngine.

    This is a high-level summary of what the robot is doing and what interactions are
    appropriate.
    """

    # Statuses for an ongoing run:

    IDLE = "idle"
    """The protocol has not been started yet.

    The robot may truly be idle, or it may be executing commands with `intent: "setup"`.
    """

    RUNNING = "running"
    """The engine is actively running the protocol."""

    PAUSED = "paused"
    """A pause has been requested. Activity is paused, or will pause soon.

    (There is currently no way to tell which.)
    """

    BLOCKED_BY_OPEN_DOOR = "blocked-by-open-door"
    """The robot's door is open. Activity is paused, or will pause soon."""

    STOP_REQUESTED = "stop-requested"
    """A stop has been requested. Activity will stop soon."""

    FINISHING = "finishing"
    """The robot is doing post-run cleanup, like homing and dropping tips."""

    # Statuses for error recovery mode:

    AWAITING_RECOVERY = "awaiting-recovery"
    """The engine is waiting for external input to recover from a nonfatal error.

    New commands with `intent: "fixit"` may be enqueued, which will run immediately.
    The run can't be paused in this state, but it can be canceled, or resumed from the
    next protocol command if recovery is complete.
    """

    AWAITING_RECOVERY_PAUSED = "awaiting-recovery-paused"
    """The engine is paused while in error recovery mode. Activity is paused, or will pause soon.

    This state is not possible to enter manually. It happens when an open door
    gets closed during error recovery.
    """

    AWAITING_RECOVERY_BLOCKED_BY_OPEN_DOOR = "awaiting-recovery-blocked-by-open-door"
    """The robot's door is open while in recovery mode. Activity is paused, or will pause soon."""

    # Terminal statuses:

    STOPPED = "stopped"
    """All activity is over; it was stopped by an explicit external request."""

    FAILED = "failed"
    """All activity is over; there was a fatal error."""

    SUCCEEDED = "succeeded"
    """All activity is over; things completed without any fatal error."""


class PostRunHardwareState(Enum):
    """State of robot gantry & motors after a stop is performed and the hardware API is reset.

    HOME_AND_STAY_ENGAGED: home the gantry and keep all motors engaged. This allows the
        robot to continue performing movement actions without re-homing
    HOME_THEN_DISENGAGE: home the gantry and then disengage motors.
        Reduces current consumption of the motors and prevents coil heating.
        Re-homing is required to re-engage the motors and resume robot movement.
    STAY_ENGAGED_IN_PLACE: do not home after the stop and keep the motors engaged.
        Keeps gantry in the same position as prior to `stop()` execution
        and allows the robot to execute movement commands without requiring to re-home first.
    DISENGAGE_IN_PLACE: disengage motors and do not home the robot
    Probable states for pipette:
        - for 1- or 8-channel:
            - HOME_AND_STAY_ENGAGED after protocol runs
            - STAY_ENGAGED_IN_PLACE after maintenance runs
        - for 96-channel:
            - HOME_THEN_DISENGAGE after protocol runs
            - DISENGAGE_IN_PLACE after maintenance runs
    """

    HOME_AND_STAY_ENGAGED = "homeAndStayEngaged"
    HOME_THEN_DISENGAGE = "homeThenDisengage"
    STAY_ENGAGED_IN_PLACE = "stayEngagedInPlace"
    DISENGAGE_IN_PLACE = "disengageInPlace"



"""Stacker-specific error codes and exceptions."""

from opentrons.drivers.asyncio.communication.errors import (
    BaseErrorCode,
    ErrorResponse,
    UnhandledGcode,
)


class EStopTriggered(ErrorResponse):
    """Raised when the estop is triggered during a move."""

    def __init__(self, port: str, response: str, command: str) -> None:
        super().__init__(port, response, command)


class MotorStallDetected(ErrorResponse):
    """Raised when a motor stall is detected."""

    def __init__(self, port: str, response: str, command: str) -> None:
        super().__init__(port, response, command)


class MotorQueueFull(ErrorResponse):
    """Raised when the motor command queue is full."""

    def __init__(self, port: str, response: str, command: str) -> None:
        super().__init__(port, response, command)


class UnexpectedLimitSwitch(ErrorResponse):
    """Raised when an unexpected limit switch is triggered."""

    def __init__(self, port: str, response: str, command: str) -> None:
        super().__init__(port, response, command)


class MotorBusy(ErrorResponse):
    """Raised when a motor is busy."""

    # TODO: differentiate between motors
    def __init__(self, port: str, response: str, command: str) -> None:
        super().__init__(port, response, command)


class StopRequested(ErrorResponse):
    """Raised when a stop is requested during a movement."""

    def __init__(self, port: str, response: str, command: str) -> None:
        super().__init__(port, response, command)


class StackerErrorCodes(BaseErrorCode):
    """Stacker-specific error codes."""

    UNHANDLED_GCODE = ("ERR003", UnhandledGcode)
    ESTOP_TRIGGERED = ("ERR006", EStopTriggered)
    MOTOR_STALL_DETECTED = ("ERR403", MotorStallDetected)
    MOTOR_QUEUE_FULL = ("ERR404", MotorQueueFull)
    UNEXPECTED_LIMIT_SWITCH = ("ERR405", UnexpectedLimitSwitch)
    X_MOTOR_BUSY = ("ERR501", MotorBusy)
    Z_MOTOR_BUSY = ("ERR502", MotorBusy)
    L_MOTOR_BUSY = ("ERR503", MotorBusy)
    STOP_REQUESTED = ("ERR504", StopRequested)



from enum import Enum

from typing_extensions import Final


class GCODE(str, Enum):
    HOME = "G28.2"
    MOVE = "G0"
    DWELL = "G4"
    CURRENT_POSITION = "M114.2"
    LIMIT_SWITCH_STATUS = "M119"
    PROBE = "G38.2"
    ABSOLUTE_COORDS = "G90"
    RELATIVE_COORDS = "G91"
    RESET_FROM_ERROR = "M999"
    PUSH_SPEED = "M120"
    POP_SPEED = "M121"
    SET_SPEED = "G0"
    STEPS_PER_MM = "M92"
    READ_INSTRUMENT_ID = "M369"
    WRITE_INSTRUMENT_ID = "M370"
    READ_INSTRUMENT_MODEL = "M371"
    WRITE_INSTRUMENT_MODEL = "M372"
    SET_MAX_SPEED = "M203.1"
    SET_CURRENT = "M907"
    DISENGAGE_MOTOR = "M18"
    HOMING_STATUS = "G28.6"
    ACCELERATION = "M204"
    WAIT = "M400"
    PIPETTE_RETRACT = "M365.3"
    PIPETTE_DEBOUNCE = "M365.2"
    PIPETTE_MAX_TRAVEL = "M365.1"
    PIPETTE_HOME = "M365.0"
    MICROSTEPPING_B_ENABLE = "M52"
    MICROSTEPPING_B_DISABLE = "M53"
    MICROSTEPPING_C_ENABLE = "M54"
    MICROSTEPPING_C_DISABLE = "M55"
    VERSION = "version"


HOMED_POSITION: Final = {
    "X": 418.0,
    "Y": 353.0,
    "Z": 218.0,
    "A": 218.0,
    "B": 19.0,
    "C": 19.0,
}

Y_BOUND_OVERRIDE: Final = 370

SMOOTHIE_COMMAND_TERMINATOR = "\r\n\r\n"

SMOOTHIE_ACK = "ok\r\nok\r\n"

PLUNGER_BACKLASH_MM = 0.3

LOW_CURRENT_Z_SPEED = 30

CURRENT_CHANGE_DELAY = 0.005

PIPETTE_READ_DELAY = 0.1

Y_SWITCH_BACK_OFF_MM = 28

Y_SWITCH_REVERSE_BACK_OFF_MM = 10

Y_BACKOFF_LOW_CURRENT = 0.8

Y_BACKOFF_SLOW_SPEED = 50

Y_RETRACT_SPEED = 8

Y_RETRACT_DISTANCE = 3

UNSTICK_DISTANCE = 1

UNSTICK_SPEED = 1

DEFAULT_AXES_SPEED = 400

XY_HOMING_SPEED = 80

HOME_SEQUENCE = ["ZABC", "X", "Y"]

AXES = "".join(HOME_SEQUENCE)

DISABLE_AXES = ""
"""Ignore these axis when sending move or home command"""

MOVEMENT_ERROR_MARGIN = 1 / 160  # Largest movement in mm for any step

SEC_PER_MIN = 60

DEFAULT_ACK_TIMEOUT = 5

DEFAULT_EXECUTE_TIMEOUT = 12000

DEFAULT_SMOOTHIE_TIMEOUT = 1

DEFAULT_MOVEMENT_TIMEOUT = 30

SMOOTHIE_BOOT_TIMEOUT = 3

DEFAULT_STABILIZE_DELAY = 0.1

DEFAULT_COMMAND_RETRIES = 3

MICROSTEPPING_GCODES = {
    "B": {
        "ENABLE": GCODE.MICROSTEPPING_B_ENABLE,
        "DISABLE": GCODE.MICROSTEPPING_B_DISABLE,
    },
    "C": {
        "ENABLE": GCODE.MICROSTEPPING_C_ENABLE,
        "DISABLE": GCODE.MICROSTEPPING_C_DISABLE,
    },
}

GCODE_ROUNDING_PRECISION = 3
"""Number of digits after the decimal point for coordinates being sent to Smoothie"""



import asyncio
import inspect
import logging
import traceback
import sys
from types import TracebackType
from typing import Any, Dict, Optional

from opentrons.drivers.smoothie_drivers.errors import SmoothieAlarm
from opentrons.protocol_api import ProtocolContext, ParameterContext
from opentrons.protocol_api._parameters import Parameters
from opentrons.protocols.execution.errors import ExceptionInProtocolError
from opentrons.protocols.types import PythonProtocol, MalformedPythonProtocolError
from opentrons.protocol_engine.types import (
    PrimitiveRunTimeParamValuesType,
    CSVRuntimeParamPaths,
)


from opentrons_shared_data.errors.exceptions import ExecutionCancelledError

MODULE_LOG = logging.getLogger(__name__)


def _runfunc_ok(run_func: Any) -> None:
    if not callable(run_func):
        raise SyntaxError("No function 'run(ctx)' defined")
    sig = inspect.Signature.from_callable(run_func)
    if not sig.parameters:
        raise SyntaxError("Function 'run()' does not take any parameters")
    if len(sig.parameters) > 1:
        for name, param in list(sig.parameters.items())[1:]:
            if param.default == inspect.Parameter.empty:
                raise SyntaxError(
                    "Function 'run{}' must be called with more than one "
                    "argument but would be called as 'run(ctx)'".format(str(sig))
                )


def _add_parameters_func_ok(add_parameters_func: Any) -> None:
    if not callable(add_parameters_func):
        raise SyntaxError("'add_parameters' must be a function.")
    sig = inspect.Signature.from_callable(add_parameters_func)
    if len(sig.parameters) != 1:
        raise SyntaxError("Function 'add_parameters' must take exactly one argument.")


def _find_protocol_error(
    tb: TracebackType | None, proto_name: str
) -> traceback.FrameSummary:
    """Return the FrameInfo for the lowest frame in the traceback from the
    protocol.
    """
    if tb is None:
        raise KeyError
    tb_info = traceback.extract_tb(tb)
    for frame in reversed(tb_info):
        if frame.filename == proto_name:
            return frame
    else:
        raise KeyError


def _raise_pretty_protocol_error(exception: Exception, filename: str) -> None:
    _, _, tb = sys.exc_info()
    try:
        frame = _find_protocol_error(tb, filename)
    except KeyError:
        # No pretty names, just raise it
        raise exception
    raise ExceptionInProtocolError(
        exception, tb, str(exception), frame.lineno
    ) from exception


def _parse_and_set_parameters(
    parameter_context: ParameterContext,
    run_time_param_overrides: Optional[PrimitiveRunTimeParamValuesType],
    run_time_param_file_overrides: Optional[CSVRuntimeParamPaths],
    new_globs: Dict[Any, Any],
    filename: str,
) -> Parameters:
    try:
        _add_parameters_func_ok(new_globs.get("add_parameters"))
    except SyntaxError as se:
        raise MalformedPythonProtocolError(str(se))
    new_globs["__param_context"] = parameter_context
    try:
        exec("add_parameters(__param_context)", new_globs)
        if run_time_param_overrides is not None:
            parameter_context.set_parameters(run_time_param_overrides)
        if run_time_param_file_overrides is not None:
            parameter_context.initialize_csv_files(run_time_param_file_overrides)
    except Exception as e:
        _raise_pretty_protocol_error(exception=e, filename=filename)
    return parameter_context.export_parameters_for_protocol()


def _get_filename(
    protocol: PythonProtocol,
) -> str:
    # TODO(mm, 2023-10-11): This coupling to opentrons.protocols.parse is fragile.
    # Can we get the correct filename directly from proto.contents?
    if protocol.filename and protocol.filename.endswith("zip"):
        # The ".zip" extension needs to match what opentrons.protocols.parse recognizes as a bundle,
        # and the "protocol.ot2.py" fallback needs to match what opentrons.protocol.py sets as the
        # AST filename.
        return "protocol.ot2.py"
    else:
        # "<protocol>" needs to match what opentrons.protocols.parse sets as the fallback
        # AST filename.
        return protocol.filename or "<protocol>"


def exec_add_parameters(
    protocol: PythonProtocol,
    parameter_context: ParameterContext,
    run_time_param_overrides: Optional[PrimitiveRunTimeParamValuesType],
    run_time_param_file_overrides: Optional[CSVRuntimeParamPaths],
) -> Optional[Parameters]:
    """Exec the add_parameters function and get the final run time parameters with overrides."""
    new_globs: Dict[Any, Any] = {}
    exec(protocol.contents, new_globs)
    filename = _get_filename(protocol)

    return (
        _parse_and_set_parameters(
            parameter_context=parameter_context,
            run_time_param_overrides=run_time_param_overrides,
            run_time_param_file_overrides=run_time_param_file_overrides,
            new_globs=new_globs,
            filename=filename,
        )
        if new_globs.get("add_parameters")
        else None
    )


def exec_run(
    proto: PythonProtocol,
    context: ProtocolContext,
    run_time_parameters_with_overrides: Optional[Parameters] = None,
) -> None:
    new_globs: Dict[Any, Any] = {}
    exec(proto.contents, new_globs)
    # If the protocol is written correctly, it will have defined a function
    # like run(context: ProtocolContext). If so, that function is now in the
    # current scope.
    filename = _get_filename(proto)
    if run_time_parameters_with_overrides:
        context._params = run_time_parameters_with_overrides

    try:
        _runfunc_ok(new_globs.get("run"))
    except SyntaxError as se:
        raise MalformedPythonProtocolError(str(se))

    new_globs["__context"] = context
    try:
        exec("run(__context)", new_globs)
    except (
        SmoothieAlarm,
        asyncio.CancelledError,
        ExecutionCancelledError,
    ):
        # this is a protocol cancel and shouldn't have special logging
        raise
    except Exception as e:
        _raise_pretty_protocol_error(exception=e, filename=filename)



"""Command models to prepare the stacker shuttle for movement."""

from __future__ import annotations

from __future__ import annotations
from typing import Literal, Union, TYPE_CHECKING
from typing_extensions import Type

from pydantic import BaseModel, Field

from .common import FlexStackerStallOrCollisionError
from opentrons_shared_data.errors.exceptions import FlexStackerStallError

from ..command import (
    AbstractCommandImpl,
    BaseCommand,
    BaseCommandCreate,
    SuccessData,
    DefinedErrorData,
)
from ...errors import ErrorOccurrence
from ...resources import ModelUtils

if TYPE_CHECKING:
    from ...state.state import StateView
    from ...execution import EquipmentHandler

PrepareShuttleCommandType = Literal["flexStacker/prepareShuttle"]


class PrepareShuttleParams(BaseModel):
    """The parameters for a PrepareShuttle command."""

    moduleId: str = Field(..., description="Unique ID of the Flex Stacker")
    ignoreLatch: bool = Field(
        default=False, description="Ignore the latch state of the shuttle"
    )


class PrepareShuttleResult(BaseModel):
    """Result data from a stacker PrepareShuttle command."""


_ExecuteReturn = Union[
    SuccessData[PrepareShuttleResult],
    DefinedErrorData[FlexStackerStallOrCollisionError],
]


class PrepareShuttleImpl(AbstractCommandImpl[PrepareShuttleParams, _ExecuteReturn]):
    """Implementation of a stacker prepare shuttle command."""

    def __init__(
        self,
        state_view: StateView,
        equipment: EquipmentHandler,
        model_utils: ModelUtils,
        **kwargs: object,
    ) -> None:
        self._state_view = state_view
        self._equipment = equipment
        self._model_utils = model_utils

    async def execute(self, params: PrepareShuttleParams) -> _ExecuteReturn:
        """Execute the stacker prepare shuttle command."""
        stacker_state = self._state_view.modules.get_flex_stacker_substate(
            params.moduleId
        )
        # Allow propagation of ModuleNotAttachedError.
        stacker_hw = self._equipment.get_module_hardware_api(stacker_state.module_id)

        try:
            if stacker_hw is not None:
                await stacker_hw.home_all(params.ignoreLatch)
        except FlexStackerStallError as e:
            return DefinedErrorData(
                public=FlexStackerStallOrCollisionError(
                    id=self._model_utils.generate_id(),
                    createdAt=self._model_utils.get_timestamp(),
                    wrappedErrors=[
                        ErrorOccurrence.from_failed(
                            id=self._model_utils.generate_id(),
                            createdAt=self._model_utils.get_timestamp(),
                            error=e,
                        )
                    ],
                ),
            )
        # TODO we should also add a check for shuttle not detected error

        return SuccessData(public=PrepareShuttleResult())


class PrepareShuttle(
    BaseCommand[PrepareShuttleParams, PrepareShuttleResult, ErrorOccurrence]
):
    """A command to prepare Flex Stacker shuttle."""

    commandType: PrepareShuttleCommandType = "flexStacker/prepareShuttle"
    params: PrepareShuttleParams
    result: PrepareShuttleResult | None = None

    _ImplementationCls: Type[PrepareShuttleImpl] = PrepareShuttleImpl


class PrepareShuttleCreate(BaseCommandCreate[PrepareShuttleParams]):
    """A request to execute a Flex Stacker PrepareShuttle command."""

    commandType: PrepareShuttleCommandType = "flexStacker/prepareShuttle"
    params: PrepareShuttleParams

    _CommandCls: Type[PrepareShuttle] = PrepareShuttle



"""Command models to close the latch of a Flex Stacker."""

from __future__ import annotations

from __future__ import annotations
from typing import Optional, Literal, TYPE_CHECKING
from typing_extensions import Type

from pydantic import BaseModel, Field

from ..command import AbstractCommandImpl, BaseCommand, BaseCommandCreate, SuccessData
from ...errors import (
    ErrorOccurrence,
)

if TYPE_CHECKING:
    from ...state.state import StateView
    from opentrons.protocol_engine.execution import EquipmentHandler

CloseLatchCommandType = Literal["flexStacker/closeLatch"]


class CloseLatchParams(BaseModel):
    """The parameters defining how a stacker should close its latch."""

    moduleId: str = Field(..., description="Unique ID of the Flex Stacker")


class CloseLatchResult(BaseModel):
    """Result data from a stacker CloseLatch command."""


class CloseLatchImpl(
    AbstractCommandImpl[CloseLatchParams, SuccessData[CloseLatchResult]]
):
    """Implementation of a stacker CloseLatch command."""

    def __init__(
        self, state_view: StateView, equipment: EquipmentHandler, **kwargs: object
    ) -> None:
        self._state_view = state_view
        self._equipment = equipment

    async def execute(self, params: CloseLatchParams) -> SuccessData[CloseLatchResult]:
        """Execute the stacker CloseLatch command."""
        stacker_state = self._state_view.modules.get_flex_stacker_substate(
            params.moduleId
        )
        stacker_hw = self._equipment.get_module_hardware_api(stacker_state.module_id)

        if stacker_hw is not None:
            await stacker_hw.close_latch()
        return SuccessData(public=CloseLatchResult())


class CloseLatch(BaseCommand[CloseLatchParams, CloseLatchResult, ErrorOccurrence]):
    """A command to CloseLatch the Flex Stacker of labware."""

    commandType: CloseLatchCommandType = "flexStacker/closeLatch"
    params: CloseLatchParams
    result: Optional[CloseLatchResult] = None

    _ImplementationCls: Type[CloseLatchImpl] = CloseLatchImpl


class CloseLatchCreate(BaseCommandCreate[CloseLatchParams]):
    """A request to execute a Flex Stacker CloseLatch command."""

    commandType: CloseLatchCommandType = "flexStacker/closeLatch"
    params: CloseLatchParams

    _CommandCls: Type[CloseLatch] = CloseLatch



"""Convert labware offset creation requests and stored elements between legacy and new."""

from opentrons_shared_data.robot.types import RobotType
from opentrons_shared_data.deck.types import DeckDefinitionV5
from .errors import (
    OffsetLocationInvalidError,
    FixtureDoesNotExistError,
)
from .types import (
    LabwareOffsetCreate,
    LegacyLabwareOffsetCreate,
    LabwareOffsetCreateInternal,
    LegacyLabwareOffsetLocation,
    LabwareOffsetLocationSequence,
    OnLabwareOffsetLocationSequenceComponent,
    OnAddressableAreaOffsetLocationSequenceComponent,
    OnModuleOffsetLocationSequenceComponent,
    ModuleModel,
)
from .resources import deck_configuration_provider


def standardize_labware_offset_create(
    request: LabwareOffsetCreate | LegacyLabwareOffsetCreate,
    robot_type: RobotType,
    deck_definition: DeckDefinitionV5,
) -> LabwareOffsetCreateInternal:
    """Turn a union of old and new labware offset create requests into a new one."""
    location_sequence, legacy_location = _locations_for_create(
        request, robot_type, deck_definition
    )
    return LabwareOffsetCreateInternal(
        definitionUri=request.definitionUri,
        locationSequence=location_sequence,
        legacyLocation=legacy_location,
        vector=request.vector,
    )


def legacy_offset_location_to_offset_location_sequence(
    location: LegacyLabwareOffsetLocation, deck_definition: DeckDefinitionV5
) -> LabwareOffsetLocationSequence:
    """Convert a legacy location to a new-style sequence."""
    sequence: LabwareOffsetLocationSequence = []
    if location.definitionUri:
        sequence.append(
            OnLabwareOffsetLocationSequenceComponent(labwareUri=location.definitionUri)
        )
    if location.moduleModel:
        sequence.append(
            OnModuleOffsetLocationSequenceComponent(moduleModel=location.moduleModel)
        )
        cutout_id = deck_configuration_provider.get_cutout_id_by_deck_slot_name(
            location.slotName
        )

        # Given a module model, try to figure out the equivalent cutout fixture.
        #
        # The Thermocycler is special. A single Thermocycler is represented in a deck
        # configuration as two separate cutout fixtures, because it spans two separate
        # cutouts. This makes it the only module whose module model string does not map
        # 1:1 with a cutout fixture ID string.
        #
        # TODO(mm, 2025-04-11): This is fragile, and the failure mode when it does the
        # wrong thing can mean labware offsets don't apply, which is pretty bad. We
        # either need a more explicit module<->cutout-fixture mapping, or we need to
        # avoid this mapping entirely.
        if (
            # Check for v2 specifically because v1 is OT-2-only and OT-2s don't have
            # modules in their deck definitions; and v3 does not exist at the time of writing.
            location.moduleModel
            == ModuleModel.THERMOCYCLER_MODULE_V2
        ):
            possible_cutout_fixture_id = "thermocyclerModuleV2Front"
        else:
            possible_cutout_fixture_id = location.moduleModel.value

        try:
            addressable_area = deck_configuration_provider.get_labware_hosting_addressable_area_name_for_cutout_and_cutout_fixture(
                cutout_id, possible_cutout_fixture_id, deck_definition
            )
            sequence.append(
                OnAddressableAreaOffsetLocationSequenceComponent(
                    addressableAreaName=addressable_area
                )
            )
        except FixtureDoesNotExistError:
            # this is an OT-2 (or this module isn't supported in the deck definition) and we should use a
            # slot addressable area name
            sequence.append(
                OnAddressableAreaOffsetLocationSequenceComponent(
                    addressableAreaName=location.slotName.value
                )
            )

    else:
        # Slight hack: we should have a more formal association here. However, since the slot
        # name is already standardized, and since the addressable areas for slots are just the
        # name of the slots, we can rely on this.
        sequence.append(
            OnAddressableAreaOffsetLocationSequenceComponent(
                addressableAreaName=location.slotName.value
            )
        )
    return sequence


def _offset_location_sequence_head_to_labware_and_module(
    location_sequence: LabwareOffsetLocationSequence,
) -> tuple[ModuleModel | None, str | None]:
    labware_uri: str | None = None
    module_model: ModuleModel | None = None
    for location in location_sequence:
        if isinstance(location, OnAddressableAreaOffsetLocationSequenceComponent):
            raise OffsetLocationInvalidError(
                "Addressable areas may only be the final element of an offset location."
            )
        elif isinstance(location, OnLabwareOffsetLocationSequenceComponent):
            if labware_uri is not None:
                # We only take the first location
                continue
            if module_model is not None:
                # Labware can't be underneath modules
                raise OffsetLocationInvalidError(
                    "Labware must not be underneath a module."
                )
            labware_uri = location.labwareUri
        elif isinstance(location, OnModuleOffsetLocationSequenceComponent):
            if module_model is not None:
                # Bad, somebody put more than one module in here
                raise OffsetLocationInvalidError(
                    "Only one module location may exist in an offset location."
                )
            module_model = location.moduleModel
        else:
            raise OffsetLocationInvalidError(
                f"Invalid location component in offset location: {repr(location)}"
            )
    return module_model, labware_uri


def _offset_location_sequence_to_legacy_offset_location(
    location_sequence: LabwareOffsetLocationSequence, deck_definition: DeckDefinitionV5
) -> LegacyLabwareOffsetLocation:
    if len(location_sequence) == 0:
        raise OffsetLocationInvalidError(
            "Offset locations must contain at least one component."
        )
    last_element = location_sequence[-1]
    if not isinstance(last_element, OnAddressableAreaOffsetLocationSequenceComponent):
        raise OffsetLocationInvalidError(
            "Offset locations must end with an addressable area."
        )
    module_model, labware_uri = _offset_location_sequence_head_to_labware_and_module(
        location_sequence[:-1]
    )
    (
        cutout_id,
        _cutout_fixtures,
    ) = deck_configuration_provider.get_potential_cutout_fixtures(
        last_element.addressableAreaName, deck_definition
    )
    slot_name = deck_configuration_provider.get_deck_slot_for_cutout_id(cutout_id)
    return LegacyLabwareOffsetLocation(
        slotName=slot_name, moduleModel=module_model, definitionUri=labware_uri
    )


def _locations_for_create(
    request: LabwareOffsetCreate | LegacyLabwareOffsetCreate,
    robot_type: RobotType,
    deck_definition: DeckDefinitionV5,
) -> tuple[LabwareOffsetLocationSequence, LegacyLabwareOffsetLocation]:
    if isinstance(request, LabwareOffsetCreate):
        return (
            request.locationSequence,
            _offset_location_sequence_to_legacy_offset_location(
                request.locationSequence, deck_definition
            ),
        )
    else:
        normalized = request.location.model_copy(
            update={
                "slotName": request.location.slotName.to_equivalent_for_robot_type(
                    robot_type
                )
            }
        )
        return (
            legacy_offset_location_to_offset_location_sequence(
                normalized, deck_definition
            ),
            normalized,
        )



"""Pick up tip command request, result, and implementation models."""

from __future__ import annotations
from opentrons_shared_data.errors import ErrorCodes
from pydantic import Field
from typing import TYPE_CHECKING, Optional, Type, Union
from typing_extensions import Literal


from ..errors import ErrorOccurrence, PickUpTipTipNotAttachedError
from ..resources import ModelUtils
from ..state import update_types
from ..types import PickUpTipWellLocation
from .pipetting_common import (
    PipetteIdMixin,
)
from .movement_common import (
    DestinationPositionResult,
    StallOrCollisionError,
    move_to_well,
)
from .command import (
    AbstractCommandImpl,
    BaseCommand,
    BaseCommandCreate,
    DefinedErrorData,
    SuccessData,
)

if TYPE_CHECKING:
    from ..state.state import StateView
    from ..execution import MovementHandler, TipHandler


PickUpTipCommandType = Literal["pickUpTip"]


class PickUpTipParams(PipetteIdMixin):
    """Payload needed to move a pipette to a specific well."""

    labwareId: str = Field(..., description="Identifier of labware to use.")
    wellName: str = Field(..., description="Name of well to use in labware.")
    wellLocation: PickUpTipWellLocation = Field(
        default_factory=PickUpTipWellLocation,
        description="Relative well location at which to pick up the tip.",
    )


class PickUpTipResult(DestinationPositionResult):
    """Result data from the execution of a PickUpTip."""

    # Tip volume has a default ONLY for parsing data from earlier versions, which did not include this in the result
    tipVolume: float = Field(
        0,
        description="Maximum volume of liquid that the picked up tip can hold, in µL.",
        ge=0,
    )

    tipLength: float = Field(
        0,
        description="The length of the tip in mm.",
        ge=0,
    )

    tipDiameter: float = Field(
        0,
        description="The diameter of the tip in mm.",
        ge=0,
    )


class TipPhysicallyMissingError(ErrorOccurrence):
    """Returned when sensors determine that no tip was physically picked up.

    That space in the tip rack is marked internally as not having any tip,
    as if the tip were consumed by a pickup.

    The pipette will act as if no tip was picked up. So, you won't be able to aspirate
    anything, and movement commands will assume there is no tip hanging off the bottom
    of the pipette.
    """

    # The thing above about marking the tips as used makes it so that
    # when the protocol is resumed and the Python Protocol API calls
    # `get_next_tip()`, we'll move on to other tips as expected.

    isDefined: bool = True
    errorType: Literal["tipPhysicallyMissing"] = "tipPhysicallyMissing"
    errorCode: str = ErrorCodes.TIP_PICKUP_FAILED.value.code
    detail: str = "No Tip Detected"


_ExecuteReturn = Union[
    SuccessData[PickUpTipResult],
    DefinedErrorData[TipPhysicallyMissingError]
    | DefinedErrorData[StallOrCollisionError],
]


class PickUpTipImplementation(AbstractCommandImpl[PickUpTipParams, _ExecuteReturn]):
    """Pick up tip command implementation."""

    def __init__(
        self,
        state_view: StateView,
        tip_handler: TipHandler,
        model_utils: ModelUtils,
        movement: MovementHandler,
        **kwargs: object,
    ) -> None:
        self._state_view = state_view
        self._tip_handler = tip_handler
        self._model_utils = model_utils
        self._movement = movement

    async def execute(
        self, params: PickUpTipParams
    ) -> Union[SuccessData[PickUpTipResult], _ExecuteReturn]:
        """Move to and pick up a tip using the requested pipette."""
        pipette_id = params.pipetteId
        labware_id = params.labwareId
        well_name = params.wellName

        tips_to_mark_as_used = self._state_view.tips.compute_tips_to_mark_as_used(
            labware_id=labware_id,
            well_name=well_name,
            nozzle_map=self._state_view.pipettes.get_nozzle_configuration(pipette_id),
        )

        well_location = self._state_view.geometry.convert_pick_up_tip_well_location(
            well_location=params.wellLocation
        )
        move_result = await move_to_well(
            movement=self._movement,
            model_utils=self._model_utils,
            pipette_id=pipette_id,
            labware_id=labware_id,
            well_name=well_name,
            well_location=well_location,
        )
        if isinstance(move_result, DefinedErrorData):
            return move_result

        try:
            tip_geometry = await self._tip_handler.pick_up_tip(
                pipette_id=pipette_id,
                labware_id=labware_id,
                well_name=well_name,
            )
        except PickUpTipTipNotAttachedError as e:
            state_update_if_false_positive = (
                update_types.StateUpdate.reduce(
                    update_types.StateUpdate(), move_result.state_update
                )
                .update_pipette_tip_state(
                    pipette_id=pipette_id,
                    tip_geometry=e.tip_geometry,
                )
                .set_fluid_empty(pipette_id=pipette_id, clean_tip=True)
                .mark_tips_as_used(
                    labware_id=labware_id, well_names=tips_to_mark_as_used
                )
            )
            state_update = (
                update_types.StateUpdate.reduce(
                    update_types.StateUpdate(), move_result.state_update
                )
                .mark_tips_as_used(
                    labware_id=labware_id, well_names=tips_to_mark_as_used
                )
                .set_fluid_unknown(pipette_id=pipette_id)
            )
            return DefinedErrorData(
                public=TipPhysicallyMissingError(
                    id=self._model_utils.generate_id(),
                    createdAt=self._model_utils.get_timestamp(),
                    wrappedErrors=[
                        ErrorOccurrence.from_failed(
                            id=self._model_utils.generate_id(),
                            createdAt=self._model_utils.get_timestamp(),
                            error=e,
                        )
                    ],
                ),
                state_update=state_update,
                state_update_if_false_positive=state_update_if_false_positive,
            )
        else:
            state_update = (
                move_result.state_update.update_pipette_tip_state(
                    pipette_id=pipette_id,
                    tip_geometry=tip_geometry,
                )
                .mark_tips_as_used(
                    labware_id=labware_id, well_names=tips_to_mark_as_used
                )
                .set_fluid_empty(pipette_id=pipette_id, clean_tip=True)
                .set_pipette_ready_to_aspirate(
                    pipette_id=pipette_id, ready_to_aspirate=True
                )
            )
            return SuccessData(
                public=PickUpTipResult(
                    tipVolume=tip_geometry.volume,
                    tipLength=tip_geometry.length,
                    tipDiameter=tip_geometry.diameter,
                    position=move_result.public.position,
                ),
                state_update=state_update,
            )


class PickUpTip(
    BaseCommand[
        PickUpTipParams,
        PickUpTipResult,
        TipPhysicallyMissingError | StallOrCollisionError,
    ]
):
    """Pick up tip command model."""

    commandType: PickUpTipCommandType = "pickUpTip"
    params: PickUpTipParams
    result: Optional[PickUpTipResult] = None

    _ImplementationCls: Type[PickUpTipImplementation] = PickUpTipImplementation


class PickUpTipCreate(BaseCommandCreate[PickUpTipParams]):
    """Pick up tip command creation request model."""

    commandType: PickUpTipCommandType = "pickUpTip"
    params: PickUpTipParams

    _CommandCls: Type[PickUpTip] = PickUpTip



from copy import deepcopy
from dataclasses import asdict
from typing import Dict, Any, List, cast, Union, Optional, TypeVar
from typing_extensions import Final
from .types import RobotConfig, CurrentDict, AxisDict

ROBOT_CONFIG_VERSION: Final = 4
PLUNGER_CURRENT_LOW = 0.05
PLUNGER_CURRENT_HIGH = 0.05

MOUNT_CURRENT_LOW = 0.1
MOUNT_CURRENT_HIGH = 0.8

X_CURRENT_LOW = 0.3
X_CURRENT_HIGH = 1.25

Y_CURRENT_LOW = 0.3
Y_CURRENT_HIGH = 1.25

XY_CURRENT_LOW_REFRESH = 0.7
MOUNT_CURRENT_HIGH_REFRESH = 0.5

Z_RETRACT_DISTANCE = 2

HIGH_CURRENT: CurrentDict = {
    "default": {
        "X": X_CURRENT_HIGH,
        "Y": Y_CURRENT_HIGH,
        "Z": MOUNT_CURRENT_HIGH_REFRESH,
        "A": MOUNT_CURRENT_HIGH_REFRESH,
        "B": PLUNGER_CURRENT_HIGH,
        "C": PLUNGER_CURRENT_HIGH,
    },
    "2.1": {
        "X": X_CURRENT_HIGH,
        "Y": Y_CURRENT_HIGH,
        "Z": MOUNT_CURRENT_HIGH,
        "A": MOUNT_CURRENT_HIGH,
        "B": PLUNGER_CURRENT_HIGH,
        "C": PLUNGER_CURRENT_HIGH,
    },
}

LOW_CURRENT: CurrentDict = {
    "default": {
        "X": XY_CURRENT_LOW_REFRESH,
        "Y": XY_CURRENT_LOW_REFRESH,
        "Z": MOUNT_CURRENT_LOW,
        "A": MOUNT_CURRENT_LOW,
        "B": PLUNGER_CURRENT_LOW,
        "C": PLUNGER_CURRENT_LOW,
    },
    "2.1": {
        "X": X_CURRENT_LOW,
        "Y": Y_CURRENT_LOW,
        "Z": MOUNT_CURRENT_LOW,
        "A": MOUNT_CURRENT_LOW,
        "B": PLUNGER_CURRENT_LOW,
        "C": PLUNGER_CURRENT_LOW,
    },
}

DEFAULT_CURRENT: CurrentDict = {
    "default": {
        "X": HIGH_CURRENT["default"]["X"],
        "Y": HIGH_CURRENT["default"]["Y"],
        "Z": HIGH_CURRENT["default"]["Z"],
        "A": HIGH_CURRENT["default"]["A"],
        "B": LOW_CURRENT["default"]["B"],
        "C": LOW_CURRENT["default"]["C"],
    },
    "2.1": {
        "X": HIGH_CURRENT["2.1"]["X"],
        "Y": HIGH_CURRENT["2.1"]["Y"],
        "Z": HIGH_CURRENT["2.1"]["Z"],
        "A": HIGH_CURRENT["2.1"]["A"],
        "B": LOW_CURRENT["2.1"]["B"],
        "C": LOW_CURRENT["2.1"]["C"],
    },
}

X_MAX_SPEED = 600
Y_MAX_SPEED = 400
Z_MAX_SPEED = 125
A_MAX_SPEED = 125
B_MAX_SPEED = 40
C_MAX_SPEED = 40

DEFAULT_MAX_SPEEDS: AxisDict = {
    "X": X_MAX_SPEED,
    "Y": Y_MAX_SPEED,
    "Z": Z_MAX_SPEED,
    "A": A_MAX_SPEED,
    "B": B_MAX_SPEED,
    "C": C_MAX_SPEED,
}

DEFAULT_CURRENT_STRING = " ".join(
    ["{}{}".format(key, value) for key, value in DEFAULT_CURRENT.items()]
)

DEFAULT_DECK_CALIBRATION_V2: List[List[float]] = [
    [1.00, 0.00, 0.00],
    [0.00, 1.00, 0.00],
    [0.00, 0.00, 1.00],
]

DEFAULT_SIMULATION_CALIBRATION: List[List[float]] = [
    [1.0, 0.0, 0.0, 0.0],
    [0.0, 1.0, 0.0, 0.0],
    [0.0, 0.0, 1.0, -25.0],
    [0.0, 0.0, 0.0, 1.0],
]

X_ACCELERATION = 3000
Y_ACCELERATION = 2000
Z_ACCELERATION = 1500
A_ACCELERATION = 1500
B_ACCELERATION = 200
C_ACCELERATION = 200

DEFAULT_ACCELERATION: Dict[str, float] = {
    "X": X_ACCELERATION,
    "Y": Y_ACCELERATION,
    "Z": Z_ACCELERATION,
    "A": A_ACCELERATION,
    "B": B_ACCELERATION,
    "C": C_ACCELERATION,
}

DEFAULT_PIPETTE_CONFIGS: Dict[str, float] = {
    "homePosition": 220,
    "stepsPerMM": 768,
    "maxTravel": 30,
}

DEFAULT_GANTRY_STEPS_PER_MM: Dict[str, float] = {
    "X": 80.00,
    "Y": 80.00,
    "Z": 400,
    "A": 400,
}


DEFAULT_MOUNT_OFFSET = [-34, 0, 0]
DEFAULT_PIPETTE_OFFSET = [0.0, 0.0, 0.0]
SERIAL_SPEED = 115200
DEFAULT_LOG_LEVEL = "INFO"

DictType = TypeVar("DictType", bound=Dict[str, Any])


def _build_dict_with_default(
    from_conf: Union[DictType, str, None], default: DictType
) -> DictType:
    if not isinstance(from_conf, dict):
        return default
    else:
        return cast(DictType, from_conf)


def _build_hw_versioned_current_dict(
    from_conf: Optional[Dict[str, Any]], default: CurrentDict
) -> CurrentDict:
    if not from_conf or not isinstance(from_conf, dict):
        return default
    # special case: if this is a valid old (i.e. not model-specific) current
    # setup, migrate it.
    if "default" not in from_conf and not (set("XYZABC") - set(from_conf.keys())):
        new_dct = deepcopy(default)
        # Because there's no case in which a machine with a more recent revision
        # than 2.1 should have a valid and edited robot config when updating
        # to this code, we should default it to 2.1 to avoid breaking other
        # robots
        new_dct["2.1"] = cast(AxisDict, from_conf)
        return new_dct
    return cast(CurrentDict, from_conf)


def build_with_defaults(robot_settings: Dict[str, Any]) -> RobotConfig:
    return RobotConfig(
        model="OT-2 Standard",
        name=robot_settings.get("name", "Ada Lovelace"),
        version=ROBOT_CONFIG_VERSION,
        gantry_steps_per_mm=_build_dict_with_default(
            robot_settings.get("steps_per_mm"), DEFAULT_GANTRY_STEPS_PER_MM
        ),
        acceleration=_build_dict_with_default(
            robot_settings.get("acceleration"), DEFAULT_ACCELERATION
        ),
        serial_speed=robot_settings.get("serial_speed", SERIAL_SPEED),
        default_current=_build_hw_versioned_current_dict(
            robot_settings.get("default_current"), DEFAULT_CURRENT
        ),
        low_current=_build_hw_versioned_current_dict(
            robot_settings.get("low_current"), LOW_CURRENT
        ),
        high_current=_build_hw_versioned_current_dict(
            robot_settings.get("high_current"), HIGH_CURRENT
        ),
        default_max_speed=robot_settings.get("default_max_speed", DEFAULT_MAX_SPEEDS),
        log_level=robot_settings.get("log_level", DEFAULT_LOG_LEVEL),
        default_pipette_configs=robot_settings.get(
            "default_pipette_configs", DEFAULT_PIPETTE_CONFIGS
        ),
        z_retract_distance=robot_settings.get("z_retract_distance", Z_RETRACT_DISTANCE),
        left_mount_offset=robot_settings.get("left_mount_offset", DEFAULT_MOUNT_OFFSET),
    )


def serialize(config: RobotConfig) -> Dict[str, Any]:
    config_dict = asdict(config)
    config_dict.pop("model", None)
    return config_dict



"""Aspirate command request, result, and implementation models."""

from __future__ import annotations
from typing import TYPE_CHECKING, Optional, Type, Union
from typing_extensions import Literal

from .pipetting_common import (
    OverpressureError,
    PipetteIdMixin,
    AspirateVolumeMixin,
    FlowRateMixin,
    BaseLiquidHandlingResult,
    aspirate_in_place,
    prepare_for_aspirate,
    DEFAULT_CORRECTION_VOLUME,
)
from .movement_common import (
    LiquidHandlingWellLocationMixin,
    DestinationPositionResult,
    StallOrCollisionError,
    move_to_well,
)
from .command import (
    AbstractCommandImpl,
    BaseCommand,
    BaseCommandCreate,
    DefinedErrorData,
    SuccessData,
)

from opentrons.hardware_control import HardwareControlAPI

from ..state.update_types import StateUpdate, CLEAR
from ..types import (
    WellLocation,
    WellOrigin,
    CurrentWell,
)

if TYPE_CHECKING:
    from ..execution import MovementHandler, PipettingHandler
    from ..resources import ModelUtils
    from ..state.state import StateView
    from ..notes import CommandNoteAdder


AspirateCommandType = Literal["aspirate"]


class AspirateParams(
    PipetteIdMixin,
    AspirateVolumeMixin,
    FlowRateMixin,
    LiquidHandlingWellLocationMixin,
):
    """Parameters required to aspirate from a specific well."""

    pass


class AspirateResult(BaseLiquidHandlingResult, DestinationPositionResult):
    """Result data from execution of an Aspirate command."""

    pass


_ExecuteReturn = Union[
    SuccessData[AspirateResult],
    DefinedErrorData[OverpressureError] | DefinedErrorData[StallOrCollisionError],
]


class AspirateImplementation(AbstractCommandImpl[AspirateParams, _ExecuteReturn]):
    """Aspirate command implementation."""

    def __init__(
        self,
        pipetting: PipettingHandler,
        state_view: StateView,
        hardware_api: HardwareControlAPI,
        movement: MovementHandler,
        command_note_adder: CommandNoteAdder,
        model_utils: ModelUtils,
        **kwargs: object,
    ) -> None:
        self._pipetting = pipetting
        self._state_view = state_view
        self._hardware_api = hardware_api
        self._movement = movement
        self._command_note_adder = command_note_adder
        self._model_utils = model_utils

    async def execute(self, params: AspirateParams) -> _ExecuteReturn:
        """Move to and aspirate from the requested well.

        Raises:
            TipNotAttachedError: if no tip is attached to the pipette.
        """
        pipette_id = params.pipetteId
        labware_id = params.labwareId
        well_name = params.wellName
        well_location = params.wellLocation

        state_update = StateUpdate()

        final_location = self._state_view.geometry.get_well_position(
            labware_id=labware_id,
            well_name=well_name,
            well_location=well_location,
            operation_volume=-params.volume,
            pipette_id=pipette_id,
        )

        ready_to_aspirate = self._pipetting.get_is_ready_to_aspirate(
            pipette_id=pipette_id
        )

        current_well = None

        if not ready_to_aspirate:
            move_result = await move_to_well(
                movement=self._movement,
                model_utils=self._model_utils,
                pipette_id=pipette_id,
                labware_id=labware_id,
                well_name=well_name,
                well_location=WellLocation(origin=WellOrigin.TOP),
            )
            state_update.append(move_result.state_update)
            if isinstance(move_result, DefinedErrorData):
                return DefinedErrorData(move_result.public, state_update=state_update)

            prepare_result = await prepare_for_aspirate(
                pipette_id=pipette_id,
                pipetting=self._pipetting,
                model_utils=self._model_utils,
                # Note that the retryLocation is the final location, inside the liquid,
                # because that's where we'd want the client to try re-aspirating if this
                # command fails and the run enters error recovery.
                location_if_error={"retryLocation": final_location},
            )
            state_update.append(prepare_result.state_update)
            if isinstance(prepare_result, DefinedErrorData):
                return DefinedErrorData(
                    public=prepare_result.public, state_update=state_update
                )

            # set our current deck location to the well now that we've made
            # an intermediate move for the "prepare for aspirate" step
            current_well = CurrentWell(
                pipette_id=pipette_id,
                labware_id=labware_id,
                well_name=well_name,
            )
        move_result = await move_to_well(
            movement=self._movement,
            model_utils=self._model_utils,
            pipette_id=pipette_id,
            labware_id=labware_id,
            well_name=well_name,
            well_location=well_location,
            current_well=current_well,
            operation_volume=-params.volume,
        )
        state_update.append(move_result.state_update)
        if isinstance(move_result, DefinedErrorData):
            return DefinedErrorData(
                public=move_result.public, state_update=state_update
            )

        aspirate_result = await aspirate_in_place(
            pipette_id=pipette_id,
            volume=params.volume,
            flow_rate=params.flowRate,
            location_if_error={
                "retryLocation": (
                    move_result.public.position.x,
                    move_result.public.position.y,
                    move_result.public.position.z,
                )
            },
            command_note_adder=self._command_note_adder,
            pipetting=self._pipetting,
            model_utils=self._model_utils,
            correction_volume=params.correctionVolume or DEFAULT_CORRECTION_VOLUME,
        )
        state_update.append(aspirate_result.state_update)
        if isinstance(aspirate_result, DefinedErrorData):
            state_update.set_liquid_operated(
                labware_id=labware_id,
                well_names=self._state_view.geometry.get_wells_covered_by_pipette_with_active_well(
                    labware_id,
                    well_name,
                    params.pipetteId,
                ),
                volume_added=CLEAR,
            )
            return DefinedErrorData(
                public=aspirate_result.public, state_update=state_update
            )

        state_update.set_liquid_operated(
            labware_id=labware_id,
            well_names=self._state_view.geometry.get_wells_covered_by_pipette_with_active_well(
                labware_id, well_name, pipette_id
            ),
            volume_added=-aspirate_result.public.volume
            * self._state_view.geometry.get_nozzles_per_well(
                labware_id,
                well_name,
                params.pipetteId,
            ),
        )

        return SuccessData(
            public=AspirateResult(
                volume=aspirate_result.public.volume,
                position=move_result.public.position,
            ),
            state_update=state_update,
        )


class Aspirate(
    BaseCommand[
        AspirateParams, AspirateResult, OverpressureError | StallOrCollisionError
    ]
):
    """Aspirate command model."""

    commandType: AspirateCommandType = "aspirate"
    params: AspirateParams
    result: Optional[AspirateResult] = None

    _ImplementationCls: Type[AspirateImplementation] = AspirateImplementation


class AspirateCreate(BaseCommandCreate[AspirateParams]):
    """Create aspirate command request model."""

    commandType: AspirateCommandType = "aspirate"
    params: AspirateParams

    _CommandCls: Type[Aspirate] = Aspirate



"""Utilities for calculating motion correctly."""
from logging import getLogger

from functools import lru_cache
from typing import Callable, Dict, Union, Optional, cast
from collections import OrderedDict

from opentrons_shared_data.robot.types import RobotType

from opentrons.types import Mount, Point
from opentrons.calibration_storage.types import AttitudeMatrix
from opentrons.util import linal

from .types import Axis, OT3Mount

log = getLogger(__name__)

# TODO: The offset_for_mount function should be defined with an overload
# set, as with other functions in this module. Unfortunately, mypy < 0.920
# has an internal crash when you mix overloads and particular kinds of decorators
# ( https://github.com/python/mypy/pull/11630 fixes it ) so instead it's defined
# with unions and therefore requires a bit of casting.
# @overload
# def offset_for_mount(
#     primary_mount: Mount,
#     left_mount_offset: Point,
#     right_mount_offset: Point,
#     gripper_mount_offset: None,
# ) -> Point:
#     ...


# @overload
# def offset_for_mount(
#     primary_mount: OT3Mount,
#     left_mount_offset: Point,
#     right_mount_offset: Point,
#     gripper_mount_offset: Point,
# ) -> Point:
#     ...

EMPTY_ORDERED_DICT = OrderedDict(
    (
        (Axis.X, 0.0),
        (Axis.Y, 0.0),
        (Axis.Z_L, 0.0),
        (Axis.Z_R, 0.0),
        (Axis.Z_G, 0.0),
        (Axis.P_L, 0.0),
        (Axis.P_R, 0.0),
        (Axis.Q, 0.0),
    )
)


@lru_cache(4)
def offset_for_mount(
    primary_mount: Union[OT3Mount, Mount],
    left_mount_offset: Point,
    right_mount_offset: Point,
    gripper_mount_offset: Optional[Point] = None,
) -> Point:
    offsets = {
        Mount.LEFT: left_mount_offset,
        Mount.RIGHT: right_mount_offset,
        OT3Mount.LEFT: left_mount_offset,
        OT3Mount.RIGHT: right_mount_offset,
        OT3Mount.GRIPPER: cast(Point, gripper_mount_offset),
    }
    return offsets[primary_mount]


def target_position_from_absolute(
    mount: Union[Mount, OT3Mount],
    abs_position: Point,
    get_critical_point: Callable[[Union[Mount, OT3Mount]], Point],
    left_mount_offset: Point,
    right_mount_offset: Point,
    gripper_mount_offset: Optional[Point] = None,
) -> "OrderedDict[Axis, float]":
    """Create a target position for all specified machine axes."""
    offset = offset_for_mount(
        mount, left_mount_offset, right_mount_offset, gripper_mount_offset
    )
    primary_cp = get_critical_point(mount)
    primary_z = Axis.by_mount(mount)

    target_position = OrderedDict(
        (
            (Axis.X, abs_position.x - offset.x - primary_cp.x),
            (Axis.Y, abs_position.y - offset.y - primary_cp.y),
            (primary_z, abs_position.z - offset.z - primary_cp.z),
        )
    )
    return target_position


def target_position_from_relative(
    mount: Union[Mount, OT3Mount],
    delta: Point,
    current_position: Dict[Axis, float],
) -> "OrderedDict[Axis, float]":
    """Create a target position for all specified machine axes."""
    primary_z = Axis.by_mount(mount)
    x_ax = Axis.X
    y_ax = Axis.Y
    target_position = OrderedDict(
        (
            (x_ax, current_position[x_ax] + delta[0]),
            (y_ax, current_position[y_ax] + delta[1]),
            (primary_z, current_position[primary_z] + delta[2]),
        )
    )
    return target_position


def target_axis_map_from_absolute(
    primary_mount: Union[OT3Mount, Mount],
    axis_map: Dict[Axis, float],
    get_critical_point: Callable[[Union[Mount, OT3Mount]], Point],
    left_mount_offset: Point,
    right_mount_offset: Point,
    gripper_mount_offset: Optional[Point] = None,
) -> "OrderedDict[Axis, float]":
    """Create an absolute target position for all specified machine axes."""
    keys_for_target_position = list(axis_map.keys())

    offset = offset_for_mount(
        primary_mount, left_mount_offset, right_mount_offset, gripper_mount_offset
    )
    primary_cp = get_critical_point(primary_mount)
    primary_z = Axis.by_mount(primary_mount)
    target_position = OrderedDict()

    if Axis.X in keys_for_target_position:
        target_position[Axis.X] = axis_map[Axis.X] - offset.x - primary_cp.x
    if Axis.Y in keys_for_target_position:
        target_position[Axis.Y] = axis_map[Axis.Y] - offset.y - primary_cp.y
    if primary_z in keys_for_target_position:
        # Since this function is intended to be used in conjunction with `API.move_axes`
        # we must leave out the carriage offset subtraction from the target position as
        # `move_axes` already does this calculation.
        target_position[primary_z] = axis_map[primary_z] - primary_cp.z

    target_position.update(
        {ax: val for ax, val in axis_map.items() if ax not in Axis.gantry_axes()}
    )
    return target_position


def target_axis_map_from_relative(
    axis_map: Dict[Axis, float],
    current_position: Dict[Axis, float],
) -> "OrderedDict[Axis, float]":
    """Create a target position for all specified machine axes."""
    target_position = OrderedDict(
        (
            (ax, current_position[ax] + axis_map[ax])
            for ax in EMPTY_ORDERED_DICT.keys()
            if ax in axis_map.keys()
        )
    )
    log.info(f"Current position {current_position} and axis map delta {axis_map}")
    log.info(f"Relative move target {target_position}")
    return target_position


def target_position_from_plunger(
    mount: Union[Mount, OT3Mount],
    delta: float,
    current_position: Dict[Axis, float],
) -> "OrderedDict[Axis, float]":
    """Create a target position for machine axes including plungers.

    Axis positions other than the plunger are identical to current
    position.
    """
    x_ax = Axis.X
    y_ax = Axis.Y
    all_axes_pos = OrderedDict(
        (
            (x_ax, current_position[x_ax]),
            (y_ax, current_position[y_ax]),
        )
    )
    plunger_pos = OrderedDict()
    z_ax = Axis.by_mount(mount)
    plunger = Axis.of_main_tool_actuator(mount)
    all_axes_pos[z_ax] = current_position[z_ax]
    plunger_pos[plunger] = delta
    all_axes_pos.update(plunger_pos)
    return all_axes_pos


def target_positions_from_plunger_tracking(
    mount: Union[Mount, OT3Mount],
    plunger_delta: float,
    z_delta: float,
    current_position: Dict[Axis, float],
) -> "OrderedDict[Axis, float]":
    """Create a target position for machine axes including plungers for dynamic liquid tracking.

    The x/y axes remain constant but the plunger and Z move to create a tracking action.

    plunger_delta: the distance the plunger should move- should be determined based on desired
    volume to aspirate/dispense.
    z_delta: the distance to move the z axis- should be determined based on volume and well geometry.
    """
    all_axes_pos = target_position_from_plunger(mount, plunger_delta, current_position)
    z_ax = Axis.by_mount(mount)
    all_axes_pos[z_ax] = current_position[z_ax] + z_delta
    return all_axes_pos


def deck_point_from_machine_point(
    machine_point: Point, attitude: AttitudeMatrix, offset: Point
) -> Point:
    return Point(
        *linal.apply_reverse(
            attitude,
            machine_point - offset,
        )
    )


def machine_point_from_deck_point(
    deck_point: Point, attitude: AttitudeMatrix, offset: Point
) -> Point:
    return Point(*linal.apply_transform(attitude, deck_point)) + offset


def machine_from_deck(
    deck_pos: Dict[Axis, float],
    attitude: AttitudeMatrix,
    offset: Point,
    robot_type: RobotType,
) -> Dict[Axis, float]:
    """Build a machine-axis position from a deck position"""
    try:
        mount_axes = (
            Axis.ot2_mount_axes()
            if robot_type == "OT-2 Standard"
            else Axis.ot3_mount_axes()
        )
        point_for_z_axis = {
            axe: Point(x=deck_pos[type(axe).X], y=deck_pos[type(axe).Y], z=pos)
            for axe, pos in deck_pos.items()
            if axe in mount_axes
        }
    except KeyError:
        raise ValueError(
            "Moves must specify either exactly an x, y, and (z or a) or none of them"
        )

    # Pre-fill the dict we’ll send to the backend with the axes we don’t
    # need to transform
    machine_pos = {
        ax: pos for ax, pos in deck_pos.items() if ax not in ax.gantry_axes()
    }

    transformed = {
        axes: machine_point_from_deck_point(deck_point, attitude, offset)
        for axes, deck_point in point_for_z_axis.items()
    }

    to_check = {}
    for axis, point in transformed.items():
        to_check[axis] = point.z
        to_check[type(axis).X] = point.x
        to_check[type(axis).Y] = point.y

    machine_pos.update({ax: pos for ax, pos in to_check.items()})
    return machine_pos


def deck_from_machine(
    machine_pos: Dict[Axis, float],
    attitude: AttitudeMatrix,
    offset: Point,
    robot_type: RobotType,
) -> Dict[Axis, float]:
    """Build a deck-abs position store from the machine's position"""
    plunger_axes: Dict[Axis, float] = {
        k: v for k, v in machine_pos.items() if k not in k.gantry_axes()
    }
    mount_axes: Dict[Axis, float]
    to_mount: Callable[[Axis], Union[Mount, OT3Mount]]
    if robot_type == "OT-2 Standard":
        mount_axes = {k: v for k, v in machine_pos.items() if k in k.ot2_mount_axes()}
        to_mount = Axis.to_ot2_mount
    else:
        mount_axes = {k: v for k, v in machine_pos.items() if k in k.ot3_mount_axes()}
        to_mount = Axis.to_ot3_mount
    deck_positions_by_mount = {
        to_mount(axis): deck_point_from_machine_point(
            Point(machine_pos[Axis.X], machine_pos[Axis.Y], value),
            attitude,
            offset,
        )
        for axis, value in mount_axes.items()
    }
    position_for_gantry = next(iter(deck_positions_by_mount.values()))
    deck_pos = {
        Axis.X: position_for_gantry[0],
        Axis.Y: position_for_gantry[1],
    }
    for mount, pos in deck_positions_by_mount.items():
        deck_pos[Axis.by_mount(mount)] = pos[2]

    deck_pos.update(plunger_axes)
    return deck_pos


def machine_vector_from_deck_vector(
    machine_vector: Point, attitude: AttitudeMatrix
) -> Point:
    """Take a vector and pass it through the attitude matrix."""
    return machine_point_from_deck_point(machine_vector, attitude, Point(0, 0, 0))



"""Helpers for commands that alter the position of labware."""

from pydantic import BaseModel, Field

from ..types import LabwareLocationSequence


class LabwareHandlingResultMixin(BaseModel):
    """A result for commands that create a labware entity."""

    labwareId: str = Field(..., description="The id of the labware.")
    locationSequence: LabwareLocationSequence | None = Field(
        None,
        description=(
            "The full location down to the deck on which this labware exists."
            " The reason this can be `null` or omitted is just backwards compatibility,"
            " for older runs and analyses. This should always be present"
            " for new runs and analyses, even for labware whose location is off-deck."
        ),
    )


class LabwarePositionResultMixin(LabwareHandlingResultMixin):
    """A result for commands that create an offsetable labware entity."""

    offsetId: str | None = Field(
        None,
        description="An ID referencing the labware offset that will apply to this labware in this location.",
    )



import hashlib
import logging
import os
import shutil
from typing import Generator, Optional, Dict, Any, TypedDict, List
from dataclasses import dataclass

from opentrons.config import CONFIG
from opentrons.system import nmcli

log = logging.getLogger(__name__)


class ConfigureArgsError(Exception):
    pass


class EapConfigItem(TypedDict):
    name: str
    displayName: str
    options: List[Dict[str, Any]]


class EapConfigShape(TypedDict):
    options: List[EapConfigItem]


EAP_CONFIG_SHAPE: EapConfigShape = {
    "options": [
        {
            "name": method.qualified_name(),
            "displayName": method.display_name(),
            "options": [
                {
                    k: v
                    for k, v in arg.items()
                    if k in ["name", "displayName", "required", "type"]
                }
                for arg in method.args()
            ],
        }
        for method in nmcli.EAP_TYPES
    ]
}


@dataclass(frozen=True)
class Key:
    directory: str
    file: str


@dataclass(frozen=True)
class AddKeyResult:
    created: bool
    key: Key


def add_key(key_file_name: str, key_contents: bytes) -> AddKeyResult:
    """
    Add a key file (for later use in EAP config) to the system.
    """
    keys_dir = CONFIG["wifi_keys_dir"]
    hasher = hashlib.sha256()
    hasher.update(key_contents)
    key_hash = hasher.hexdigest()
    if key_hash in os.listdir(keys_dir):
        files = os.listdir(os.path.join(keys_dir, key_hash))
        if files:
            return AddKeyResult(
                created=False, key=Key(directory=key_hash, file=files[0])
            )
        else:
            log.warning("Key directory with nothing in it: {}".format(key_hash))
            os.rmdir(os.path.join(keys_dir, key_hash))

    key_hash_path = os.path.join(keys_dir, key_hash)
    os.mkdir(key_hash_path)
    with open(os.path.join(key_hash_path, os.path.basename(key_file_name)), "wb") as f:
        f.write(key_contents)
    return AddKeyResult(created=True, key=Key(directory=key_hash, file=key_file_name))


def list_keys() -> Generator[Key, None, None]:
    """
    List wifi keys known to the system.

    :return: A generator yielding Key objects
    """
    keys_dir = CONFIG["wifi_keys_dir"]
    # TODO(mc, 2018-10-24): add last modified info to keys for sort purposes
    for path in os.listdir(keys_dir):
        full_path = os.path.join(keys_dir, path)
        if os.path.isdir(full_path):
            in_path = os.listdir(full_path)
            if len(in_path) > 1:
                log.warning("Garbage in key dir for key {}".format(path))
            yield Key(directory=path, file=in_path[0])
        else:
            log.warning("Garbage in wifi keys dir: {}".format(full_path))


def remove_key(requested_hash: str) -> Optional[str]:
    """
    Try to delete key file

    :param requested_hash: The hash to delete
    :return: The name of the deleted file or None if not found
    """
    keys_dir = CONFIG["wifi_keys_dir"]
    available_keys = os.listdir(keys_dir)
    if requested_hash not in available_keys:
        return None
    key_path = os.path.join(keys_dir, requested_hash)
    name = os.listdir(key_path)[0]
    shutil.rmtree(key_path)
    return name


def get_key_file(key: str) -> str:
    """
    Get the full path of a key file

    :param key: The key to look for
    :return: the path
    """
    keys_dir = CONFIG["wifi_keys_dir"]
    available_keys = os.listdir(keys_dir)
    if key not in available_keys:
        raise ConfigureArgsError(f"Key ID {key} is not valid on the system")
    files_in_dir = os.listdir(os.path.join(keys_dir, key))
    if len(files_in_dir) > 1:
        raise OSError(f"Key ID {key} has multiple files, try deleting and re-uploading")
    return os.path.join(keys_dir, key, files_in_dir[0])


def _eap_check_no_extra_args(config: Dict[str, Any], options: Any) -> None:
    # options is an Any because the type annotation for EAP_CONFIG_SHAPE itself
    # can’t quite express the type properly because of the inference from the
    # dict annotation.
    """Check for args that are not required for this method (to aid debugging)
    ``config`` should be the user config.
    ``options`` should be the options sub-member for the eap method.

    Before this method is called, the validity of the 'eapType' key should be
    established.
    """
    arg_names = [k for k in config.keys() if k != "eapType"]
    valid_names = [o["name"] for o in options]
    for an in arg_names:
        if an not in valid_names:
            raise ConfigureArgsError(
                "Option {} is not valid for EAP method {}".format(an, config["eapType"])
            )


def _eap_check_option_ok(opt: Dict[str, str], config: Dict[str, Any]) -> None:
    """
    Check that a given EAP option is in the user config (if required)
    and, if specified, is the right type.

    ``opt`` should be an options dict from EAP_CONFIG_SHAPE.
    ``config`` should be the user config dict.

    Before this method is called, the validity of the eapType key should be
    established.
    """
    if opt["name"] not in config:
        if opt["required"]:
            raise ConfigureArgsError(
                "Required argument {} for eap method {} not present".format(
                    opt["displayName"], config["eapType"]
                )
            )
        else:
            return
    name = opt["name"]
    o_type = opt["type"]
    arg = config[name]
    if name in config:
        if o_type in ("string", "password") and not isinstance(arg, str):
            raise ConfigureArgsError("Option {} should be a str".format(name))
        elif o_type == "file" and not isinstance(arg, str):
            raise ConfigureArgsError("Option {} must be a str".format(name))


def eap_check_config(eap_config: Dict[str, Any]) -> Dict[str, Any]:
    """Check the eap specific args, and replace values where needed."""
    eap_type = eap_config.get("eapType")
    for method in EAP_CONFIG_SHAPE["options"]:
        if method["name"] == eap_type:
            options = method["options"]
            break
    else:
        raise ConfigureArgsError("EAP method {} is not valid".format(eap_type))

    _eap_check_no_extra_args(eap_config, options)

    for opt in options:
        _eap_check_option_ok(opt, eap_config)
        if opt["type"] == "file" and opt["name"] in eap_config:
            # Special work for file: rewrite from key id to path
            eap_config[opt["name"]] = get_key_file(eap_config[opt["name"]])
    return eap_config



"""Configure nozzle layout command request, result, and implementation models."""
from __future__ import annotations
from opentrons.protocol_engine.state.update_types import StateUpdate
from pydantic import BaseModel
from typing import TYPE_CHECKING, Optional, Type, Union
from typing_extensions import Literal

from .pipetting_common import (
    PipetteIdMixin,
)
from .command import AbstractCommandImpl, BaseCommand, BaseCommandCreate, SuccessData
from ..errors.error_occurrence import ErrorOccurrence
from ..types import (
    AllNozzleLayoutConfiguration,
    SingleNozzleLayoutConfiguration,
    RowNozzleLayoutConfiguration,
    ColumnNozzleLayoutConfiguration,
    QuadrantNozzleLayoutConfiguration,
)

if TYPE_CHECKING:
    from ..execution import EquipmentHandler, TipHandler


ConfigureNozzleLayoutCommandType = Literal["configureNozzleLayout"]


class ConfigureNozzleLayoutParams(PipetteIdMixin):
    """Parameters required to configure the nozzle layout for a specific pipette."""

    configurationParams: Union[
        AllNozzleLayoutConfiguration,
        SingleNozzleLayoutConfiguration,
        RowNozzleLayoutConfiguration,
        ColumnNozzleLayoutConfiguration,
        QuadrantNozzleLayoutConfiguration,
    ]


class ConfigureNozzleLayoutResult(BaseModel):
    """Result data from execution of an configureNozzleLayout command."""

    pass


class ConfigureNozzleLayoutImplementation(
    AbstractCommandImpl[
        ConfigureNozzleLayoutParams,
        SuccessData[ConfigureNozzleLayoutResult],
    ]
):
    """Configure nozzle layout command implementation."""

    def __init__(
        self, equipment: EquipmentHandler, tip_handler: TipHandler, **kwargs: object
    ) -> None:
        self._equipment = equipment
        self._tip_handler = tip_handler

    async def execute(
        self, params: ConfigureNozzleLayoutParams
    ) -> SuccessData[ConfigureNozzleLayoutResult]:
        """Check that requested pipette can support the requested nozzle layout."""
        primary_nozzle = params.configurationParams.model_dump().get("primaryNozzle")
        front_right_nozzle = params.configurationParams.model_dump().get(
            "frontRightNozzle"
        )
        back_left_nozzle = params.configurationParams.model_dump().get("backLeftNozzle")
        nozzle_params = await self._tip_handler.available_for_nozzle_layout(
            pipette_id=params.pipetteId,
            style=params.configurationParams.style,
            primary_nozzle=primary_nozzle,
            front_right_nozzle=front_right_nozzle,
            back_left_nozzle=back_left_nozzle,
        )

        nozzle_map = await self._equipment.configure_nozzle_layout(
            pipette_id=params.pipetteId,
            **nozzle_params,
        )

        update_state = StateUpdate()
        update_state.update_pipette_nozzle(
            pipette_id=params.pipetteId, nozzle_map=nozzle_map
        )

        return SuccessData(
            public=ConfigureNozzleLayoutResult(),
            state_update=update_state,
        )


class ConfigureNozzleLayout(
    BaseCommand[
        ConfigureNozzleLayoutParams, ConfigureNozzleLayoutResult, ErrorOccurrence
    ]
):
    """Configure nozzle layout command model."""

    commandType: ConfigureNozzleLayoutCommandType = "configureNozzleLayout"
    params: ConfigureNozzleLayoutParams
    result: Optional[ConfigureNozzleLayoutResult] = None

    _ImplementationCls: Type[
        ConfigureNozzleLayoutImplementation
    ] = ConfigureNozzleLayoutImplementation


class ConfigureNozzleLayoutCreate(BaseCommandCreate[ConfigureNozzleLayoutParams]):
    """Configure nozzle layout creation request model."""

    commandType: ConfigureNozzleLayoutCommandType = "configureNozzleLayout"
    params: ConfigureNozzleLayoutParams

    _CommandCls: Type[ConfigureNozzleLayout] = ConfigureNozzleLayout



import functools
import inspect
from contextlib import contextmanager
from typing import Any, Callable, Iterator, Optional, TypeVar, cast
from uuid import uuid4

from opentrons.legacy_broker import LegacyBroker

from .types import (
    COMMAND as COMMAND_TOPIC,
    Command as CommandPayload,
    CommandMessage,
    MessageSequenceId,
)


class CommandPublisher:
    """An object with a `Broker` dependency used to publish commands."""

    def __init__(self, broker: Optional[LegacyBroker]) -> None:
        """Initialize the publisher with a Broker."""
        self._broker = broker or LegacyBroker()

    @property
    def broker(self) -> LegacyBroker:
        """Get the publisher's Broker."""
        return self._broker

    @broker.setter
    def broker(self, broker: LegacyBroker) -> None:
        """Set the publisher's Broker."""
        self._broker = broker


CommandPayloadCreator = Callable[..., CommandPayload]
"""A function that creates a Command dictionary message."""

FuncT = TypeVar("FuncT", bound=Callable[..., Any])
"""A function wrapped by the @publish decorator."""


def publish(command: CommandPayloadCreator) -> Callable[[FuncT], FuncT]:
    """Publish messages before and after the decorated function has run."""

    def _decorator(func: FuncT) -> FuncT:
        @functools.wraps(func)
        def _decorated(*args: Any, **kwargs: Any) -> Any:
            """Use the args passed to wrapped `func` to build the message payload.

            1. Inspect signature of `func` and bind arguments to signature to map
               argument names to called (and/or default) values.
            2. Inspect signature of `command` and map values from `func` call to
               argument names expected by `command`.
            3. Map `self` argument of `func` to the `instrument` argument of `command`,
               where applicable.
            4. Construct the command payload and publish it using `publish_context`
            5. Return the value of calling `func` with `*args` and `**kwargs`
            """

            broker = getattr(args[0], "broker", None)

            assert isinstance(
                broker, LegacyBroker
            ), "Only methods of CommandPublisher classes should be decorated."

            func_sig = _inspect_signature(func)
            bound_func_args = func_sig.bind(*args, **kwargs)
            bound_func_args.apply_defaults()
            func_args = bound_func_args.arguments

            message_creator_sig = _inspect_signature(command)
            message_creator_arg_names = set(message_creator_sig.parameters.keys())
            message_creator_args = {
                n: func_args[n] for n in message_creator_arg_names if n in func_args
            }

            # TODO (artyom, 20170927): we are doing this to be able to use
            # the decorator in Instrument class methods, in which case
            # self is effectively an instrument.
            # To narrow the scope of this hack, we are checking if the
            # command is expecting instrument first.
            # We are also checking if call arguments have 'self' and
            # don't have instruments specified, in which case
            # instruments should take precedence.
            if (
                "instrument" in message_creator_arg_names
                and "instrument" not in message_creator_args
                and "self" in func_args
            ):
                message_creator_args["instrument"] = func_args["self"]

            command_message = command(**message_creator_args)

            with publish_context(broker=broker, command=command_message):
                return func(*args, **kwargs)

        return cast(FuncT, _decorated)

    return _decorator


@contextmanager
def publish_context(broker: LegacyBroker, command: CommandPayload) -> Iterator[None]:
    """Publish messages before and after the `with` block has run.

    If an `error` is raised in the `with` block, it will be published in the "after"
    message and re-raised.
    """
    message_id = str(uuid4())
    _do_publish(broker=broker, message_id=message_id, command=command, when="before")

    try:
        yield
    except Exception as error:
        _do_publish(
            broker=broker,
            message_id=message_id,
            command=command,
            when="after",
            error=error,
        )
        raise
    else:
        _do_publish(broker=broker, message_id=message_id, command=command, when="after")


@functools.lru_cache(maxsize=None)
def _inspect_signature(func: Callable[..., Any]) -> inspect.Signature:
    """Inspect function signatures, memoized because it is called very often."""
    return inspect.signature(func)


def _do_publish(
    broker: LegacyBroker,
    message_id: str,
    command: CommandPayload,
    when: MessageSequenceId,
    error: Optional[Exception] = None,
) -> None:
    """Publish a command to the broker from the decorator or ContextManager."""
    name = command["name"]
    payload = command["payload"]
    message: CommandMessage = {  # type: ignore[assignment, misc]
        "$": when,
        "id": message_id,
        "name": name,
        "payload": payload,
        "error": error,
    }

    if when == "before":
        payload_str = ", ".join(f"{k}: {v}" for k, v in payload.items() if k != "text")
        broker.logger.info(f"{name}: {payload_str}")

    broker.publish(topic=COMMAND_TOPIC, message=message)
